# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for Python code execution."""

import inspect
import unittest
from langfun.core.coding.python import execution
import pyglove as pg


class EvaluateTest(unittest.TestCase):

  def test_with_context(self):
    with execution.context(x=1, y=0):
      with execution.context(x=2, z=2):
        self.assertEqual(
            execution.evaluate(
                """
                p = x + y + z
                """,
                # Override value from the context.
                global_vars=dict(z=3),
                outputs_intermediate=True,
            ),
            dict(p=2 + 0 + 3, __result__=2 + 0 + 3, __stdout__=''),
        )

  def test_basics(self):
    self.assertEqual(
        execution.evaluate(
            """
            x = 1
            y = x + 1
            print(y)
            z = x + y
            """,
            outputs_intermediate=True,
        ),
        dict(x=1, y=2, z=3, __result__=3, __stdout__='2\n'),
    )
    self.assertEqual(
        execution.evaluate(
            """
            x = 1
            y = x + 1
            print(y)
            z = x + y
            """,
        ),
        3,
    )
    with self.assertRaisesRegex(execution.CodeError, 'ValueError'):
      execution.evaluate(
          """
          def foo():
            raise ValueError("intentional error")
          foo()
          """,
          permission=execution.CodePermission.ALL
      )

  def test_class_def(self):
    ret = execution.evaluate(
        """
        class A(pg.Object):
          x: int
          y: int
          def __call__(self):
            return self.x + self.y
        """,
        permission=execution.CodePermission.ALL,
        global_vars=dict(pg=pg),
        outputs_intermediate=True,
    )
    self.assertEqual(list(ret.keys()), ['A', '__result__', '__stdout__'])
    self.assertTrue(issubclass(ret['A'], pg.Object))
    self.assertIs(ret['__result__'], ret['A'])
    self.assertEqual(ret['__stdout__'], '')

  def test_function_def(self):
    ret = execution.evaluate(
        """
        def foo(x, y):
          return x + y

        def bar(z):
          return z + foo(z, z)
        """,
        permission=execution.CodePermission.ALL,
        outputs_intermediate=True,
    )
    self.assertEqual(
        list(ret.keys()), ['foo', 'bar', '__result__', '__stdout__']
    )
    self.assertTrue(inspect.isfunction(ret['foo']))
    self.assertTrue(inspect.isfunction(ret['bar']))
    self.assertIs(ret['__result__'], ret['bar'])

  def test_function_def_and_call(self):
    code = (
        """
        def foo(x, y):
          return x + y

        def bar(z):
          print(f'z is {z}')
          return z + foo(z, z)

        bar(1)
        """
    )
    ret = execution.evaluate(
        code,
        permission=execution.CodePermission.ALL,
        outputs_intermediate=True,
    )
    self.assertEqual(
        list(ret.keys()), ['foo', 'bar', '__result__', '__stdout__']
    )
    self.assertEqual(ret['__result__'], 3)
    ret = execution.evaluate(
        code,
        permission=execution.CodePermission.ALL,
        returns_stdout=True,
    )
    self.assertEqual(ret, 'z is 1\n')

  def test_complex(self):
    ret = execution.evaluate(
        """
        class A(pg.Object):
          x: int
          y: int
          def __call__(self, z):
            return self.x + self.y + z

        def foo(x, y):
          return x + y
        k = A(1, 2)
        k(foo(3, 4))
        """,
        permission=execution.CodePermission.ALL,
        global_vars=dict(pg=pg),
        outputs_intermediate=True,
    )
    self.assertEqual(
        list(ret.keys()), ['A', 'foo', 'k', '__result__', '__stdout__']
    )
    self.assertTrue(issubclass(ret['A'], pg.Object))
    self.assertTrue(inspect.isfunction(ret['foo']))
    self.assertIsInstance(ret['k'], pg.Object)
    self.assertEqual(ret['__result__'], 10)

  def test_run_with_error(self):
    with self.assertRaisesRegex(
        execution.CodeError, 'NameError: name .* is not defined'
    ):
      execution.evaluate(
          """
          x = 1
          y = x + z
          """,
          permission=execution.CodePermission.ALL,
      )
    with self.assertRaisesRegex(execution.CodeError, 'ValueError'):
      execution.evaluate(
          'raise ValueError()', permission=execution.CodePermission.ALL
      )


class Foo(pg.Object):
  x: int
  y: int


class RunTest(unittest.TestCase):

  def test_run_without_sandboxing(self):
    self.assertEqual(
        execution.run(
            'x + y',
            global_vars=dict(x=1, y=2),
            sandbox=False,
        ),
        3,
    )

  def test_run_with_sandboxing(self):
    self.assertEqual(
        execution.run(
            'x + y',
            global_vars=dict(x=1, y=2),
            sandbox=True,
        ),
        3,
    )

  def test_run_with_automatic_sandboxing(self):
    self.assertEqual(
        execution.run(
            'x + y',
            global_vars=dict(x=1, y=2),
        ),
        3,
    )

    r = execution.run(
        inspect.cleandoc("""
            def foo(x, y):
              return x + y

            class A(pg.Object):
              x: str
            """),
        global_vars=dict(pg=pg),
        outputs_intermediate=True,
    )
    self.assertTrue(inspect.isfunction(r['foo']))
    self.assertTrue(inspect.isclass(r['A']))


if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for structured mapping example."""

import inspect
from typing import Any
import unittest

import langfun.core as lf
from langfun.core.structured import mapping
import pyglove as pg


class MappingErrorTest(unittest.TestCase):

  def test_format(self):
    error = mapping.MappingError(
        lf.AIMessage('hi'), ValueError('Cannot parse message.')
    )
    self.assertEqual(
        pg.decolor(str(error)),
        'ValueError: Cannot parse message.\n\n[LM Response]\nhi',
    )
    self.assertEqual(
        pg.decolor(error.format(include_lm_response=False)),
        'ValueError: Cannot parse message.',
    )


class MappingExampleTest(unittest.TestCase):

  def test_basics(self):
    m = mapping.MappingExample('Compute 1 + 1', '2')
    self.assertEqual(m.schema_repr(), '')

    m = mapping.MappingExample('Compute 1 + 1', '2', schema=int)
    self.assertEqual(m.schema_repr('python'), 'int')
    self.assertEqual(m.schema_repr('json'), '{"result": int}')

  def test_str(self):
    self.assertEqual(
        str(
            mapping.MappingExample(
                input='1 + 1 = 2',
                output=2,
                context='Give the answer.',
                schema=int,
            )
        ),
        inspect.cleandoc("""
            \x1b[1m[CONTEXT]
            \x1b[0m\x1b[35mGive the answer.\x1b[0m

            \x1b[1m[INPUT]
            \x1b[0m\x1b[32m1 + 1 = 2\x1b[0m

            \x1b[1m[SCHEMA]
            \x1b[0m\x1b[31mint\x1b[0m

            \x1b[1m[OUTPUT]
            \x1b[0m\x1b[34m```python
            2
            ```\x1b[0m
            """),
    )

  def test_str_no_context(self):
    self.assertEqual(
        str(mapping.MappingExample('1 + 1 = 2', 2, schema=int)),
        inspect.cleandoc("""
            \x1b[1m[INPUT]
            \x1b[0m\x1b[32m1 + 1 = 2\x1b[0m

            \x1b[1m[SCHEMA]
            \x1b[0m\x1b[31mint\x1b[0m

            \x1b[1m[OUTPUT]
            \x1b[0m\x1b[34m```python
            2
            ```\x1b[0m
            """),
    )

  def test_str_no_schema(self):
    self.assertEqual(
        str(mapping.MappingExample('1 + 1 = 2', 2, context='Give the answer.')),
        inspect.cleandoc("""
            \x1b[1m[CONTEXT]
            \x1b[0m\x1b[35mGive the answer.\x1b[0m

            \x1b[1m[INPUT]
            \x1b[0m\x1b[32m1 + 1 = 2\x1b[0m

            \x1b[1m[OUTPUT]
            \x1b[0m\x1b[34m```python
            2
            ```\x1b[0m
            """),
    )

  def test_str_no_output(self):
    self.assertEqual(
        str(
            mapping.MappingExample(
                '1 + 1 = 2',
                schema=int,
                context='Give the answer.',
            )
        ),
        inspect.cleandoc("""
            \x1b[1m[CONTEXT]
            \x1b[0m\x1b[35mGive the answer.\x1b[0m

            \x1b[1m[INPUT]
            \x1b[0m\x1b[32m1 + 1 = 2\x1b[0m

            \x1b[1m[SCHEMA]
            \x1b[0m\x1b[31mint\x1b[0m
            """),
    )

  def test_str_with_metadata(self):
    self.assertEqual(
        str(
            mapping.MappingExample(
                '1 + 1 = 2',
                schema=int,
                context='Give the answer.',
                metadata={'foo': 'bar'},
            )
        ),
        inspect.cleandoc("""
            \x1b[1m[CONTEXT]
            \x1b[0m\x1b[35mGive the answer.\x1b[0m

            \x1b[1m[INPUT]
            \x1b[0m\x1b[32m1 + 1 = 2\x1b[0m

            \x1b[1m[SCHEMA]
            \x1b[0m\x1b[31mint\x1b[0m

            \x1b[1m[METADATA]
            \x1b[0m\x1b[36m{
              foo = 'bar'
            }\x1b[0m
            """),
    )

  def test_serialization(self):
    example = mapping.MappingExample(
        'the answer is 2', 2, int, context='compute 1 + 1'
    )
    self.assertTrue(
        pg.eq(pg.from_json_str(example.to_json_str()), example)
    )

  def assert_html_content(self, html, expected):
    expected = inspect.cleandoc(expected).strip()
    actual = html.content.strip()
    if actual != expected:
      print(actual)
    self.assertEqual(actual, expected)

  def test_html(self):

    class Answer(pg.Object):
      answer: int

    class Addition(lf.Template):
      """Template Addition.

      {{x}} + {{y}} = ?
      """
      x: Any
      y: Any

    example = mapping.MappingExample(
        input=Addition(x=1, y=2),
        schema=Answer,
        context='compute 1 + 1',
        output=Answer(answer=3),
        metadata={'foo': 'bar'},
    )
    self.assert_html_content(
        example.to_html(
            enable_summary_tooltip=False,
            extra_flags=dict(
                include_message_metadata=False
            )
        ),
        """
        <details open class="pyglove mapping-example"><summary><div class="summary-title">MappingExample(...)</div></summary><div class="complex-value mapping-example"><details open class="pyglove user-message lf-message"><summary><div class="summary-name lf-message">input<span class="tooltip lf-message">input</span></div><div class="summary-title lf-message">UserMessage(...)</div></summary><div class="complex_value"><div class="message-tags"><span>rendered</span></div><div class="message-text">1 + 2 = ?</div></div></details><details open class="pyglove answer lf-example-output"><summary><div class="summary-name lf-example-output">output<span class="tooltip lf-example-output">output</span></div><div class="summary-title lf-example-output">Answer(...)</div></summary><div class="complex-value answer"><details open class="pyglove int"><summary><div class="summary-name">answer<span class="tooltip">output.answer</span></div><div class="summary-title">int</div></summary><span class="simple-value int">3</span></details></div></details><details open class="pyglove str"><summary><div class="summary-name">context<span class="tooltip">context</span></div><div class="summary-title">str</div></summary><span class="simple-value str">&#x27;compute 1 + 1&#x27;</span></details><details open class="pyglove schema lf-example-schema"><summary><div class="summary-name lf-example-schema">schema<span class="tooltip lf-example-schema">schema</span></div><div class="summary-title lf-example-schema">Schema(...)</div></summary><div class="lf-schema-definition">Answer

        ```python
        class Answer:
          answer: int
        ```</div></details><details open class="pyglove dict lf-example-metadata"><summary><div class="summary-name lf-example-metadata">metadata<span class="tooltip lf-example-metadata">metadata</span></div><div class="summary-title lf-example-metadata">Dict(...)</div></summary><div class="complex-value dict"><details open class="pyglove str"><summary><div class="summary-name">foo<span class="tooltip">metadata.foo</span></div><div class="summary-title">str</div></summary><span class="simple-value str">&#x27;bar&#x27;</span></details></div></details></div></details>
        """
    )

    example = mapping.MappingExample(
        input=Addition(x=1, y=2),
        output=Answer(answer=3),
    )
    self.assert_html_content(
        example.to_html(
            enable_summary_tooltip=False,
            extra_flags=dict(
                include_message_metadata=False
            )
        ),
        """
        <details open class="pyglove mapping-example"><summary><div class="summary-title">MappingExample(...)</div></summary><div class="complex-value mapping-example"><details open class="pyglove user-message lf-message"><summary><div class="summary-name lf-message">input<span class="tooltip lf-message">input</span></div><div class="summary-title lf-message">UserMessage(...)</div></summary><div class="complex_value"><div class="message-tags"><span>rendered</span></div><div class="message-text">1 + 2 = ?</div></div></details><details open class="pyglove answer lf-example-output"><summary><div class="summary-name lf-example-output">output<span class="tooltip lf-example-output">output</span></div><div class="summary-title lf-example-output">Answer(...)</div></summary><div class="complex-value answer"><details open class="pyglove int"><summary><div class="summary-name">answer<span class="tooltip">output.answer</span></div><div class="summary-title">int</div></summary><span class="simple-value int">3</span></details></div></details><details open class="pyglove contextual-attribute lf-example-schema"><summary><div class="summary-name lf-example-schema">schema<span class="tooltip lf-example-schema">schema</span></div><div class="summary-title lf-example-schema">ContextualAttribute(...)</div></summary><span class="simple-value none-type">None</span></details><details open class="pyglove dict lf-example-metadata"><summary><div class="summary-name lf-example-metadata">metadata<span class="tooltip lf-example-metadata">metadata</span></div><div class="summary-title lf-example-metadata">Dict(...)</div></summary><div class="complex-value dict"><span class="empty-container"></span></div></details></div></details>
        """
    )


if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""langfun subscription framework."""

import abc
import collections
import dataclasses
import functools
import inspect
import typing
from typing import Any, Callable, Generic, Iterator, Sequence, Type, TypeVar, Union
import weakref


SenderType = TypeVar('SenderType')


@dataclasses.dataclass
class Event(Generic[SenderType]):
  sender: SenderType


EventType = TypeVar('EventType')


class EventHandler(Generic[EventType], metaclass=abc.ABCMeta):
  """Interface for event subscriber."""

  @classmethod
  @functools.cache
  def event_type(cls) -> Type[Event[Any]]:
    """Returns acceptable event type."""
    return _get_generic_arg(cls, EventHandler)

  @classmethod
  @functools.cache
  def sender_type(cls) -> Type[Any]:
    """Returns acceptable sender type."""
    return _get_generic_arg(cls.event_type(), Event)

  @classmethod
  def accepts(cls, event: Event[Any]) -> bool:
    """Returns True if current event handler class can accepts an event."""
    return isinstance(event, cls.event_type())

  @abc.abstractmethod
  def on_event(self, event: EventType) -> None:
    """Handles an event."""


class _EventManager:
  """Event manager."""

  def __init__(self):
    self._sender_registry = weakref.WeakKeyDictionary()
    self._sender_type_registry: dict[
        Type[Any], list[EventHandler[Any]]  # Sender type.
    ] = collections.defaultdict(list)

  def _map_sender_subscriber(
      self,
      func: Callable[
          [
              Union[Any, Type[Any], None],  # Sender object, type or None.
              EventHandler[Any],
          ],  # Subscriber (event handler).
          None,
      ],
      sender_or_senders: Union[
          Any, Type[Any], Sequence[Union[Any, Type[Any]]], None
      ],
      subscriber_or_subscribers: Union[
          EventHandler[Any], Sequence[EventHandler[Any]]
      ],
  ) -> None:
    """Maps the product of senders and subscribers as the input to a function."""
    if isinstance(subscriber_or_subscribers, (tuple, list)):
      subscriber_list = list(subscriber_or_subscribers)
    else:
      subscriber_list = [subscriber_or_subscribers]

    if isinstance(sender_or_senders, (tuple, list)):
      senders = list(sender_or_senders)
    else:
      senders = [sender_or_senders]

    for subscriber in subscriber_list:
      for sender in senders:
        func(subscriber, sender)

  def _sender_info(
      self, sender: Union[Any, Type[Any], None]
  ) -> tuple[
      Any,  # Sender or None.
      Type[Any] | None,  # Sender type or None.
      list[EventHandler[Any]],  # Subscribers.
  ]:
    """Returns the sender id, sender type and current subscribers."""
    if inspect.isclass(sender):
      return None, sender, self._sender_type_registry[sender]
    elif sender is not None:
      sender_entry = self._sender_registry.get(sender, None)
      if sender_entry is None:
        sender_entry = []
        self._sender_registry[sender] = sender_entry
      return sender, type(sender), sender_entry
    else:
      return None, None, []

  def _subscribe(
      self,
      subscriber: EventHandler[Any],
      sender: Union[Any, Type[Any], None] = None,
  ) -> None:
    """Subscribes a subscriber to a sender."""
    if not isinstance(subscriber, EventHandler):
      raise TypeError(f'{subscriber!r} is not an event handler.')

    sender, sender_type, subscriber_list = self._sender_info(sender)
    subscriber_sender_type = subscriber.sender_type()

    if sender_type is None:
      sender_type = subscriber_sender_type
      subscriber_list = self._sender_type_registry[subscriber_sender_type]
    elif not issubclass(sender_type, subscriber_sender_type):
      raise TypeError(
          f'{sender_type!r} is not the expected sender type '
          f'({subscriber_sender_type!r}) for subscriber '
          f'{subscriber.__class__.__name__!r}'
      )
    assert subscriber_list is not None
    subscriber_list.append(subscriber)

  def _unsubscribe(
      self,
      subscriber: EventHandler[Any],
      sender: Union[Any, Type[Any], None] = None,
  ) -> None:
    """Unsubscribes a subscriber from a sender."""
    # Unsubscribe.
    unsubscribed = 0
    sender, sender_type, _ = self._sender_info(sender)

    if sender is not None:
      # Unsubscribe a single sender.
      if sender in self._sender_registry:
        unsubscribed += _remove_list_items(
            self._sender_registry[sender], lambda x: x is subscriber
        )
    else:
      # Unsubscribe all subscriptions for the subscriber.
      subscriber_sender_type = subscriber.sender_type()
      sender_type = sender_type or subscriber_sender_type

      # Unsubscribe object-based subscriptions.
      for subscriber_list in self._sender_registry.values():
        unsubscribed += _remove_list_items(
            subscriber_list, lambda x: x is subscriber
        )

      # Unsubscribe type-based subscriptions.
      for st, subscriber_list in self._sender_type_registry.items():
        if issubclass(st, sender_type):
          unsubscribed += _remove_list_items(
              subscriber_list, lambda x: x is subscriber
          )

    if unsubscribed == 0:
      raise ValueError(
          f'There is no subscription for {subscriber!r} to unsubscribe.'
      )

  def subscribe(
      self,
      subscriber: Union[EventHandler[Any], Sequence[EventHandler[Any]]],
      sender: Union[
          Any, Type[Any], Sequence[Union[Any, Type[Any]]], None
      ] = None,
  ) -> None:
    """Subscribes one or a list subscribers to one or a list of senders."""
    return self._map_sender_subscriber(
        self._subscribe,
        sender_or_senders=sender,
        subscriber_or_subscribers=subscriber,
    )

  def unsubscribe(
      self,
      subscriber: Union[EventHandler[Any], Sequence[EventHandler[Any]]],
      sender: Union[
          Any, Type[Any], Sequence[Union[Any, Type[Any]]], None
      ] = None,
  ) -> None:
    """Unsubscribes one or a list subscribers from one or a list of senders."""
    return self._map_sender_subscriber(
        self._unsubscribe,
        sender_or_senders=sender,
        subscriber_or_subscribers=subscriber,
    )

  def emit(self, event: Event[Any]) -> None:
    """Emits an event."""
    for subscriber in self.subscribers(event.sender):
      if subscriber.accepts(event):
        subscriber.on_event(event)

  def subscribers(self, sender: Any | Type[Any]) -> Iterator[EventHandler[Any]]:
    """Iterates the subscribers of a sender."""
    visited = set()
    sender, sender_type, _ = self._sender_info(sender)

    # Yield instance-level subscribers.
    if sender is not None and sender in self._sender_registry:
      for subscriber in self._sender_registry[sender]:
        if id(subscriber) not in visited:
          yield subscriber
          visited.add(id(subscriber))

    # Yield type_level subscribers.
    for registered_type, subscriber_list in self._sender_type_registry.items():
      if isinstance(sender, registered_type) or (
          sender is None and issubclass(sender_type, registered_type)
      ):
        for subscriber in subscriber_list:
          if id(subscriber) not in visited:
            yield subscriber
            visited.add(id(subscriber))

  def subscriptions(self, subscriber: EventHandler[Any]) -> Iterator[Any]:
    """Returns all subscriptions of a subscriber."""
    # Yield subscriptions based on individual senders.
    for sender, subscriber_list in self._sender_registry.items():
      if any(x is subscriber for x in subscriber_list):
        yield sender

    # Yield subscriptions based on sender types.
    for sender_type, subscribers_list in self._sender_type_registry.items():
      if any(x is subscriber for x in subscribers_list):
        yield sender_type

  def clear(self) -> None:
    """Clear all subscriptions."""
    self._sender_registry.clear()
    self._sender_type_registry.clear()


_event_manager = _EventManager()


def subscribe(
    subscriber: Union[EventHandler[Any], Sequence[EventHandler[Any]]],
    sender: Union[Any, Type[Any], Sequence[Union[Any, Type[Any]]], None] = None,
) -> None:
  """Subscribes one or a list subscribers to one or a list of senders.

  Args:
    subscriber: An event subscriber or a list of event handlers whose event
      sender matches with the sender.
    sender: One or a list of sender object or sender type, whose emitted events
      should be received. If None, the expecting sender type from the subscriber
      will be used.
  """
  _event_manager.subscribe(subscriber, sender)


def unsubscribe(
    subscriber: Union[EventHandler[Any], Sequence[EventHandler[Any]]],
    sender: Union[Any, Type[Any], Sequence[Union[Any, Type[Any]]], None] = None,
) -> None:
  """Unsubscribes one or a list subscribers from one or a list of senders.

  Args:
    subscriber: An event subscriber or a list of event handlers whose event
      sender matches with the sender.
    sender: One or a list of sender object or sender type, whose emitted events
      should be received. If None, the expecting sender type from the subscriber
      will be used.
  """
  _event_manager.unsubscribe(subscriber, sender)


def emit(event: Event[Any]) -> None:
  """Emits an event."""
  _event_manager.emit(event)


def subscribers(sender: Any | Type[Any]) -> Iterator[EventHandler[Any]]:
  """Iterates the subscribers of a sender or sender type."""
  return _event_manager.subscribers(sender)


def subscriptions(subscriber: EventHandler[Any]) -> Iterator[Any]:
  """Iterates the subscriptions of a subscriber."""
  return _event_manager.subscriptions(subscriber)


def clear_subscriptions() -> None:
  """Clear all subscriptions."""
  _event_manager.clear()


def _get_generic_arg(cls, generic_cls):
  """Get the first type argument from a generic type or subtype."""
  for orig_base in getattr(cls, '__orig_bases__', ()):
    if typing.get_origin(orig_base) is generic_cls:
      return typing.get_args(orig_base)[0]
  raise TypeError(
      f'Cannot get type argument from generic class {generic_cls}` '
      f'type from {cls}'
  )


def _remove_list_items(list_values, pred) -> int:
  """Remove list items based on predicate."""
  indices_to_remove = []
  for i, v in enumerate(list_values):
    if pred(v):
      indices_to_remove.append(i)
  for i in reversed(indices_to_remove):
    list_values.pop(i)
  return len(indices_to_remove)



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for langfun.core.templates.Conversation."""

import inspect
import unittest
import langfun.core as lf
from langfun.core.templates.conversation import Conversation


class QuestionCounter(lf.LanguageModel):

  def _on_bound(self):
    super()._on_bound()
    self._num_call = 0

  def _sample(self, prompts):
    del prompts
    self._num_call += 1
    return [lf.LMSamplingResult([lf.LMSample(f'Response {self._num_call}.')])]


class ConversationTest(unittest.TestCase):

  def test_render_input_only(self):
    c = Conversation(next_input='How are you?')
    self.assertEqual(
        c.render(),
        'How are you?')

  def test_preamble_only(self):
    c = Conversation(
        'You are a helpful and joyful chat bot. Let us chat.',
        'How are you?')
    self.assertEqual(
        c.render(),
        inspect.cleandoc("""
          You are a helpful and joyful chat bot. Let us chat.

          User: How are you?
        """))

  def test_with_conversation_context(self):
    c = Conversation(
        preamble='You are a helpful and joyful chat bot. Let us chat.',
        conversation_context=lf.LangFunc(
            """
            {% for message in conversation_history %}
            {{message.sender}}: {{message.text}}
            {%- endfor %}
            """,
            conversation_history=[
                lf.UserMessage('Hello'),
                lf.AIMessage('Hi, how can I help you?'),
            ]
        ),
        next_input=lf.LangFunc(
            """{{message.sender}}: {{message.text}}""",
            message=lf.UserMessage('What is your name?'),
        ),
    )

    self.assertEqual(
        c.render(),
        inspect.cleandoc("""
          You are a helpful and joyful chat bot. Let us chat.
          User: Hello
          AI: Hi, how can I help you?
          User: What is your name?
        """))

  def test_call(self):
    c = Conversation(
        lm=QuestionCounter(),
        role='Agent',
        preamble="You are a helpful and joyful AI bot. Now let's chat.",
    )
    # First round.
    self.assertEqual(
        c.prompt.render(input_message='Hello'),
        inspect.cleandoc("""
        You are a helpful and joyful AI bot. Now let's chat.

        User: Hello
        """),
    )
    self.assertEqual(c(input_message='Hello'), 'Response 1.')

    # Second round.
    self.assertEqual(
        c.prompt.render(input_message='How are you?'),
        inspect.cleandoc("""
        You are a helpful and joyful AI bot. Now let's chat.
        User: Hello
        Agent: Response 1.
        User: How are you?
        """),
    )
    self.assertEqual(c(input_message='How are you?'), 'Response 2.')

    # Third round.
    self.assertEqual(
        c.prompt.render(input_message='Okay, bye.'),
        inspect.cleandoc("""
        You are a helpful and joyful AI bot. Now let's chat.
        User: Hello
        Agent: Response 1.
        User: How are you?
        Agent: Response 2.
        User: Okay, bye.
        """),
    )
    self.assertEqual(c(input_message='Okay, bye.'), 'Response 3.')

    # A new round of conversation.
    c.reset()
    self.assertEqual(
        c.prompt.render(input_message='Hi'),
        inspect.cleandoc("""
        You are a helpful and joyful AI bot. Now let's chat.

        User: Hi
        """),
    )


if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Template test."""
import inspect
from typing import Any
import unittest

from langfun.core import component
from langfun.core import message as message_lib
from langfun.core import modality
from langfun.core import subscription
from langfun.core.template import Template
from langfun.core.template import TemplateRenderEvent
import pyglove as pg


class BasicTest(unittest.TestCase):

  def test_rebind(self):
    l = Template('Hello')
    self.assertEqual(l._variables, set())
    l.rebind(template_str='Hello {{x}}')
    self.assertEqual(l._variables, set(['x']))

  def test_attrs(self):
    l = Template('Hello {{x}}', x=1)
    self.assertEqual(l.template_str, 'Hello {{x}}')
    self.assertTrue(l.clean)
    self.assertEqual(l.x, 1)

  def test_render_output(self):
    l = Template('Hello {{x}}', x=[1])
    self.assertIsNone(l.render_output)
    self.assertEqual(l.render(), 'Hello `[1]`')
    self.assertEqual(l.render_output, 'Hello `[1]`')

  def test_str(self):
    self.assertEqual(str(Template('Hello')), 'Hello')
    self.assertEqual(str(Template('Hello {{x}}')), 'Hello {{x}}')
    self.assertEqual(str(Template('Hello {{x}}', x=1)), 'Hello 1')

  def test_hash(self):
    l = Template('Hello')
    self.assertEqual(hash(l), hash(l))
    self.assertNotEqual(hash(l), hash(l.clone()))

  def test_eq(self):
    self.assertEqual(Template('Hello'), 'Hello')
    self.assertNotEqual(Template('Hello'), 'Hello\n')

    self.assertEqual(Template('Hello {{x}}', x='World'), 'Hello World')
    self.assertNotEqual(
        Template('Hello {{x}}', x='World'), Template('Hello {{x}}', x='World')
    )

    self.assertTrue(
        pg.eq(
            Template('Hello {{x}}', x='World'),
            Template('Hello {{x}}', x='World'),
        )
    )

    self.assertTrue(
        pg.ne(
            Template('Hello {{x}}', x='World'),
            Template('Hello {{x}}', x='Word'),
        )
    )

  def test_custom_typing(self):
    class Foo(component.Component):
      x: str | None = None
      y: int | None = None
      z: str
      p: component.Component | None = None

    d = Foo(z='bar')
    # String template can be assigned to str.
    d.x = Template('Hello, {{y}}')

    d.y = 1
    d.z = Template('Bye, {{y}}')

    # String template can be assigned to Component.
    d.p = Template('Again {{x}}')

    self.assertEqual(d.x.render(), 'Hello, 1')
    self.assertEqual(d.z.render(), 'Bye, 1')
    self.assertEqual(d.p.render(), 'Again Hello, 1')

  def test_raw_text(self):
    self.assertEqual(
        Template(
            '{{a}}' + Template.raw_str('\n{{d}}, {%x%}\n') + '{{b}}',
            a='hi', b=1
        ).render().text,
        'hi\n{{d}}, {%x%}\n1'
    )

  def test_from_raw_str(self):
    self.assertEqual(
        Template.from_raw_str('\n{{d}}, {%x%}\n').render().text,
        '\n{{d}}, {%x%}\n'
    )


class DefinitionTest(unittest.TestCase):

  def test_subclassing(self):

    class MyPrompt(Template):
      template_str = 'Hello {{x}}, {{y}} and {{z}}'
      x: int = 1
      p: str

      @property
      def z(self):
        return self.p + '!'

    # `p` is required.
    with self.assertRaisesRegex(TypeError, '.* missing 1 required argument'):
      MyPrompt()()

    l = MyPrompt(p='abc')
    self.assertEqual(l.render(y=2), 'Hello 1, 2 and abc!')

  def test_subclassing_with_docstr(self):

    class MyPrompt(Template):
      """My prompt.

      Hello {{x}} and {{y}}
      """

      x: int = 1

    l = MyPrompt()
    self.assertEqual(
        MyPrompt.__schema__.get_field('template_str').default_value,
        'Hello {{x}} and {{y}}',
    )
    self.assertEqual(l.render(y=2), 'Hello 1 and 2')

  def test_subclassing_with_no_template_sign_docstr(self):

    class MyPrompt(Template):
      """My prompt.

      This is a longer version of docstr.

      (THIS IS NOT A TEMPLATE)
      """

      x: int = 1

    self.assertEqual(
        MyPrompt.__schema__.get_field('template_str').default_value,
        pg.MISSING_VALUE,
    )
    with self.assertRaisesRegex(TypeError, '.* missing 1 required argument'):
      MyPrompt(y=2)()

  def test_bad_template(self):
    with self.assertRaisesRegex(ValueError, 'Bad template string.*'):
      Template('{{x=1')


class VarsTest(unittest.TestCase):

  def assert_missing_vars(self, t: Template, missing_vars: set[str]):
    self.assertEqual(t.missing_vars, missing_vars)

  def test_missing_vars(self):
    self.assert_missing_vars(Template('Hello'), set())
    self.assert_missing_vars(Template('Hello {{x}}', x=1), set())
    self.assert_missing_vars(Template('Hello {{x}}'), set(['x']))
    self.assert_missing_vars(
        Template('Hello {{x}}', x=Template('{{y}}')), set(['y'])
    )

  def test_vars(self):
    class A(Template):
      """A.

      Hello {{x}}{{y}}{{z}}
      """

      x: int

    class B(Template):
      """B.

      hi {{x}}{{p}}{{q}}
      """

      p: str = 'foo'

    b = B(q=Template('There {{i}}{{j}}{{x}}', i=0))
    a = A(x=1, y=b)

    # Test all direct referred variables.
    self.assertEqual(a.vars(), set(['x', 'y', 'z']))
    self.assertEqual(b.vars(), set(['x', 'p', 'q']))
    self.assertEqual(b.q.vars(), set(['x', 'i', 'j']))

    # Test direct referred variables that are specified.
    self.assertEqual(a.vars(specified=True), set(['x', 'y']))
    self.assertEqual(b.vars(specified=True), set(['x', 'p', 'q']))
    self.assertEqual(b.q.vars(specified=True), set(['x', 'i']))

    # Test direct referred variables that are not specified.
    self.assertEqual(a.vars(specified=False), set(['z']))
    self.assertEqual(b.vars(specified=False), set())
    self.assertEqual(b.q.vars(specified=False), set(['j']))

    # Test all referred variables in the closure.
    self.assertEqual(
        a.vars(closure=True), set(['x', 'y', 'z', 'p', 'q', 'i', 'j'])
    )
    self.assertEqual(b.vars(closure=True), set(['x', 'q', 'p', 'i', 'j']))
    self.assertEqual(b.q.vars(closure=True), set(['x', 'i', 'j']))

    # Test all leaf variables.
    self.assertEqual(
        a.vars(closure=True, leaf=True), set(['x', 'z', 'p', 'i', 'j'])
    )
    self.assertEqual(b.vars(closure=True, leaf=True), set(['x', 'p', 'i', 'j']))
    self.assertEqual(b.q.vars(closure=True, leaf=True), set(['x', 'i', 'j']))

    # Test all non-leaf variables.
    self.assertEqual(a.vars(closure=True, leaf=False), set(['y', 'q']))
    self.assertEqual(b.vars(closure=True, leaf=False), set(['q']))
    self.assertEqual(b.q.vars(closure=True, leaf=False), set())

    # Test missing variables.
    self.assertEqual(a.missing_vars, set(['z', 'j']))
    self.assertEqual(b.missing_vars, set(['j']))
    self.assertEqual(b.q.missing_vars, set(['j']))


class RenderTest(unittest.TestCase):

  def test_clean(self):
    l = Template('\n Hello\n ')
    self.assertEqual(l.render(), 'Hello')

    l = Template("""
        Hello
         {{foo}}
           {{bar}}
        """)
    self.assertEqual(l.render(foo='a', bar='b'), 'Hello\n a\n   b')

  def test_no_clean(self):
    l = Template('\n Hello\n ', clean=False)
    self.assertEqual(l.render(), '\n Hello\n ')

  def test_constant_template(self):
    l = Template('Hello')
    self.assertEqual(l.render(), 'Hello')
    self.assertEqual(l.natural_language_format(), 'Hello')

  def test_render_without_call_args(self):
    l = Template(
        'How {{x}}',
        x=Template('are {{y}}', y=Template('you {{n}}')),
        n=1,
    )
    v = l.render()
    self.assertEqual(v, 'How are you 1')
    self.assertIs(v.x, l.x)
    self.assertEqual(v.x.render_output, 'are you 1')
    self.assertIs(v.x.y, l.x.y)
    self.assertEqual(v.x.y.render_output, 'you 1')
    self.assertEqual(v.x.y.n, 1)

  def test_render_with_call_args(self):
    l = Template('Hello {{x}} and {{y}}', x=1, y=Template('{{z}}'), z=3)
    v = l.render(x=2)
    self.assertEqual(v, 'Hello 2 and 3')
    self.assertEqual(v.x, 2)
    self.assertEqual(v.y, '3')
    self.assertEqual(v.y.z, 3)

  def test_render_cache(self):
    class DynamicContent(Template):
      """Random number.

      {{_counter}}
      """

      def _on_bound(self):
        super()._on_bound()
        self._counter = 0

      def render(self, **kwargs):
        self._counter += 1
        return super().render(**kwargs)

    l = Template('{{x}} + {{x}} =', x=DynamicContent())
    self.assertEqual(l.render(), '1 + 1 =')

    l = Template('{{x}} + {{x.render()}} =', x=DynamicContent())
    self.assertEqual(l.render(), '1 + 2 =')

  def test_render_with_modality(self):
    class CustomModality(modality.Modality):
      content: str

      def to_bytes(self):
        return self.content.encode()

    self.assertEqual(
        Template(
            'This is {{ x }} and {{ a }}', x=1, a=CustomModality('foo')
        ).render(),
        'This is 1 and <<[[a]]>>',
    )

  def test_render_with_default(self):

    class Foo(Template):
      """Foo.

      This is {{x}}
      """

    f = Foo(template_str='!{{DEFAULT}}!', x=1)
    self.assertEqual(f.DEFAULT.x, 1)
    self.assertEqual(
        f.render(), '!This is 1!'
    )

    class Bar(Template):
      """Bar.

      {{preamble}}
      {{prompt}}
      """

      preamble: Template = Template('You are a chat bot.')
      prompt: Template = Template('User: hi! {{name}}')

    b = Bar(
        preamble=Template('<h1>{{DEFAULT}}</h1>'),
        prompt=Template('<h2>{{DEFAULT}}</h2>'),
        name='Tom',
    )
    # Test variable access.
    self.assertEqual(
        b.render(),
        inspect.cleandoc("""
            <h1>You are a chat bot.</h1>
            <h2>User: hi! Tom</h2>
            """),
    )

    with self.assertRaisesRegex(ValueError, '`{{ DEFAULT }}` cannot be used'):

      class Baz(Template):  # pylint: disable=unused-variable
        """Baz.

        {{DEFAULT}}
        """

    with self.assertRaisesRegex(
        ValueError, 'The template neither has a default `template_str` nor'
    ):
      Template('{{DEFAULT}}').render()

    d = pg.Dict(x=Template('{{DEFAULT}}'))
    with self.assertRaisesRegex(
        ValueError, 'does not have a default value'
    ):
      _ = d.x.DEFAULT

    class Tes(pg.Object):
      x: str | None = None

    t = Tes(x=Template('{{DEFAULT}}'))
    with self.assertRaisesRegex(
        ValueError, 'is not a `lf.Template` object or str'
    ):
      _ = t.x.DEFAULT

  def test_bad_render(self):
    with self.assertRaises(ValueError):
      Template('Hello {{x}}').render(allow_partial=False)

  def assert_partial(self, t: Template, expected_text: str):
    self.assertEqual(t.render(allow_partial=True), expected_text)

  def test_partial_rendering(self):
    self.assert_partial(Template('Hello {{x}} {{y}}'), 'Hello {{x}} {{y}}')

    self.assert_partial(Template('Hello {{x}} {{y}}', y=1), 'Hello {{x}} 1')

    self.assert_partial(
        Template('Hello {{x[0].y.z()}}'), 'Hello {{x[0].y.z()}}'
    )

    self.assert_partial(
        Template("Hello {{f(1, 2, a='foo')}}"), "Hello {{f(1, 2, a='foo')}}"
    )

    # Test arithmetic operations.
    self.assert_partial(Template('Hello {{ -x }}'), 'Hello {{-x}}')

    self.assert_partial(Template('Hello {{ x + 1 }}'), 'Hello {{x + 1}}')

    self.assert_partial(Template('Hello {{ 1 + x }}'), 'Hello {{1 + x}}')

    self.assert_partial(Template('Hello {{ x - 1 }}'), 'Hello {{x - 1}}')

    self.assert_partial(Template('Hello {{ 1 - x }}'), 'Hello {{1 - x}}')

    self.assert_partial(Template('Hello {{ x * 2 }}'), 'Hello {{x * 2}}')

    self.assert_partial(Template('Hello {{ 2 * x }}'), 'Hello {{2 * x}}')

    self.assert_partial(Template('Hello {{ x / 2 }}'), 'Hello {{x / 2}}')

    self.assert_partial(Template('Hello {{ 2 / x }}'), 'Hello {{2 / x}}')

    self.assert_partial(Template('Hello {{ x // 2 }}'), 'Hello {{x // 2}}')

    self.assert_partial(Template('Hello {{ 2 // x }}'), 'Hello {{2 // x}}')

    self.assert_partial(Template('Hello {{ x ** 2 }}'), 'Hello {{x ** 2}}')

    self.assert_partial(Template('Hello {{ 2 ** x }}'), 'Hello {{2 ** x}}')

    self.assert_partial(Template('Hello {{ x % 2 }}'), 'Hello {{x % 2}}')

    self.assert_partial(Template('Hello {{ 2 % x }}'), 'Hello {{2 % x}}')

    # Test logic operations.
    self.assert_partial(Template('Hello {{ x == 1 }}'), 'Hello {{x == 1}}')

    self.assert_partial(Template('Hello {{ x != 1 }}'), 'Hello {{x != 1}}')

    self.assert_partial(Template('Hello {{ x > y }}'), 'Hello {{x > y}}')

    self.assert_partial(Template('Hello {{ x >= y }}'), 'Hello {{x >= y}}')

    self.assert_partial(Template('Hello {{ x <= y }}'), 'Hello {{x <= y}}')

    self.assert_partial(Template('Hello {{ x < y }}'), 'Hello {{x < y}}')

    # Test list format.
    self.assert_partial(
        Template("""
            {%- for example in examples %}
            {{ example -}}
            {% endfor %}"""),
        inspect.cleandoc("""
            {{examples.item0}}
            {{examples.item1}}
            {{examples.item2}}
            """),
    )

    self.assert_partial(
        Template("""
            {%- for k in examples.keys() %}
            {{ k -}}
            {% endfor %}"""),
        inspect.cleandoc("""
            {{examples.key0}}
            {{examples.key1}}
            {{examples.key2}}
            """),
    )

    self.assert_partial(
        Template("""
            {%- for v in examples.values() %}
            {{ v -}}
            {% endfor %}"""),
        inspect.cleandoc("""
            {{examples.value0}}
            {{examples.value1}}
            {{examples.value2}}
            """),
    )

    self.assert_partial(
        Template("""
            {%- for k, v in examples.items() %}
            {{ k }}: {{ v -}}
            {% endfor %}"""),
        inspect.cleandoc("""
            {{examples.key0}}: {{examples.value0}}
            {{examples.key1}}: {{examples.value1}}
            {{examples.key2}}: {{examples.value2}}
            """),
    )

    # Test len.
    self.assert_partial(Template('Hello {{len(x)}}'), 'Hello {{len(x)}}')

  def test_additional_metadata(self):
    t = Template('hi', metadata_weights=1.0, y=2)
    self.assertEqual(t.render(), message_lib.UserMessage('hi', weights=1.0))

    t = Template('hi')
    with component.context(metadata_weights=1.0, y=2):
      self.assertEqual(t.render(), message_lib.UserMessage('hi', weights=1.0))


class TemplateRenderEventTest(unittest.TestCase):

  def test_render_event(self):
    l = Template(
        'Subject: {{subject}}', subject=Template('The science of {{name}}')
    )

    render_events = []
    render_stacks = []

    class RenderEventHandler(subscription.EventHandler[TemplateRenderEvent]):

      def on_event(self, event: TemplateRenderEvent):
        render_events.append(event.output)
        render_stacks.append(event.render_stack)

    x = RenderEventHandler()
    subscription.subscribe(x)
    l.render(name='rocket')
    self.assertEqual(
        render_events,
        ['The science of rocket', 'Subject: The science of rocket'],
    )
    self.assertEqual(render_stacks, [[l], []])
    subscription.unsubscribe(x)
    subscription.subscribe(x, l.subject)

    render_events[:] = []
    render_stacks[:] = []
    l.render(name='ballon')
    self.assertEqual(render_events, ['The science of ballon'])
    self.assertEqual(render_stacks, [[l]])


class HtmlTest(unittest.TestCase):

  def test_html(self):

    class Foo(Template):
      """Template Foo.

      {{x}} + {{y}} = ?
      """
      x: Any
      y: Any

    class Bar(Template):
      """Template Bar.

      {{y}} + {{z}}
      """
      y: Any

    self.assertIn(
        inspect.cleandoc(
            """
            /* Langfun Template styles. */
            .template-str {
                padding: 10px;
                margin: 10px 5px 10px 5px;
                font-style: italic;
                font-size: 1.1em;
                white-space: pre-wrap;
                border: 1px solid #EEE;
                border-radius: 5px;
                background-color: #EEE;
                color: #cc2986;
            }
            """
        ),
        Foo(x=1, y=2).to_html().style_section,
    )
    self.assertIn(
        'template-str',
        Foo(x=Bar('{{y}} + {{z}}'), y=1).to_html(
            enable_summary_tooltip=False,
        ).content,
    )
    self.assertIn(
        'template-str',
        Foo(x=Bar('{{y}} + {{z}}'), y=1).to_html(
            enable_summary_tooltip=False,
            collapse_level=0,
            key_style='label',
        ).content,
    )


if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""LLM-based class generation."""

import typing
from typing import Any, Type
import langfun.core as lf
from langfun.core.coding.python import correction
from langfun.core.structured import mapping
import pyglove as pg


class GenerateClass(mapping.Mapping):
  """Python class generation."""

  input_title = 'GENERATION_CONTEXT'
  context_title = 'CLASS_NAME'
  output_title = 'OUTPUT_CLASS'
  permission = pg.coding.CodePermission.ALL

  preamble: lf.Template = lf.Template("""
      Help generate a class based on the last {{ context_title }} and {{ input_title }}.

      Instructions:
      - Use `Object` as the base class for all generated classes
      - Create auxillary classes for composition if needed.
      - Use Python type annotation for declaraing fields:
        (e.g. bool, str, int, float, Optional[str], List[int], Union[str, int])
      - Do not use types that need import.
      - Avoid self-referential types. e.g:
        ```
        class Node(Object):
          children: list[Node]
        ```
      - Do not generate methods.
      """)

  def parse_result(self, lm_output: lf.Message) -> Type[Any]:
    output_vars, final_code = correction.run_with_correction(
        lm_output.text,
        global_vars=self.allowed_annotation_types,
        sandbox=False,
        max_attempts=self.autofix,
        lm=self.autofix_lm,
        returns_code=True,
        outputs_intermediate=True,
    )
    class_name = self.context
    cls = output_vars.get(class_name, None)
    if cls is None:
      raise pg.coding.CodeError(
          final_code,
          TypeError(f'Class {class_name} is absent from LLM output.'),
      )
    return cls

  @property
  def allowed_annotation_types(self):
    return dict(
        pg=pg,
        Any=typing.Any,
        Object=pg.Object,
        List=typing.List,
        Dict=typing.Tuple,
        Tuple=typing.Tuple,
        Sequence=typing.Sequence,
        Optional=typing.Optional,
        Union=typing.Union,
    )


def generate_class(
    name: str,
    prompt: str | pg.Symbolic,
    *,
    lm: lf.LanguageModel | None = None,
    examples: list[mapping.MappingExample] | None = None,
    returns_message: bool = False,
    skip_lm: bool = False,
    **kwargs,
) -> Type[Any] | lf.Message:
  """Generate a class with specified name based on the prompt.

  Example:
    ```
    trip_cls = lf.classgen(
        'Trip',
        'A trip plan to visit {{ city }}, city='San Francisco',
        lm=lf.llms.GeminiPro()
    )
    ```

  Args:
    name: Class name to be generated.
    prompt: A str (may contain {{}} as template) as natural language input, or a
      `pg.Symbolic` object as structured input as prompt to LLM.
    lm: The language model to use. If not specified, the language model from
      `lf.context` context manager will be used.
    examples: An optional list of fewshot examples for helping class generation.
      If None, a default single shot example will be used. Use
      `lf.structured.classgen_example` to generate example.
    returns_message: If True, returns `lf.Message` as the output, instead of
      returning the structured `message.result`.
    skip_lm: If True, returns the rendered prompt as a UserMessage object.
      otherwise return the LLM response based on the rendered prompt.
    **kwargs: Template variables passed to `prompt` and keyword arguments passed
      to `lf.structured.GenerateClass`.

  Returns:
    Generated class.

  Raises:
    CodeError: if generation failed.
  """
  if isinstance(prompt, str):
    prompt = lf.Template(prompt, **kwargs)
  elif isinstance(prompt, lf.Template):
    prompt = prompt.rebind(**kwargs, raise_on_no_change=False)

  if isinstance(prompt, lf.Template):
    prompt = prompt.render(lm=lm)

  call_kwargs = dict(skip_lm=skip_lm)
  if lm is not None:
    call_kwargs['lm'] = lm
  message = GenerateClass(
      input=prompt,
      context=name,
      examples=examples or default_classgen_examples(),
      **kwargs,
  )(**call_kwargs)
  return message if returns_message else message.result


def classgen_example(
    prompt: str | pg.Symbolic, cls: Type[Any]
) -> mapping.MappingExample:
  """Creates a class generation example."""
  if isinstance(prompt, lf.Template):
    prompt = prompt.render()
  return mapping.MappingExample(
      input=prompt,
      context=cls.__name__,
      output=cls,
  )


def default_classgen_examples() -> list[mapping.MappingExample]:
  """Default examples for class generation."""

  class Step(pg.Object):
    description: str
    output: float

  class Solution(pg.Object):
    steps: list[Step]  # pytype: disable=invalid-annotation
    result: float

  return [
      classgen_example(
          'How to evaluate an arithmetic expression?',
          Solution,
      )
  ]



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import unittest
import langfun.core as lf
from langfun.core.llms import fake
from langfun.core.structured import scoring
import pyglove as pg


class Answer(pg.Object):
  result: int


class ScoringTest(unittest.TestCase):

  def test_bad_call(self):
    with self.assertRaisesRegex(ValueError, '`completions` must not be empty'):
      scoring.score('hi', [])

    with self.assertRaisesRegex(
        ValueError, '`schema` cannot be inferred from completions'
    ):
      scoring.score('hi', [1, 'b'])

    with self.assertRaisesRegex(ValueError, '`lm` must be specified'):
      scoring.score('hi', [1, 2])

    with self.assertRaisesRegex(
        ValueError,
        'Scoring on object fields using `pg.oneof` must share the same prompt',
    ):
      scoring.score(
          ['1 + 1=', '2 + 3='],
          [Answer(pg.oneof([1, 2, 3]))],
          lm=fake.Echo(),
      )

  def test_score(self):
    self.assertEqual(scoring.score('hi', [1, 2], lm=fake.Echo()), [0.0, -1.0])

  def test_ascore(self):
    with lf.context(lm=fake.Echo()):
      self.assertEqual(
          asyncio.run(scoring.ascore('hi', [1, 2], lm=fake.Echo())),
          [0.0, -1.0]
      )

  def test_score_on_field_values(self):
    self.assertEqual(
        scoring.score(
            '1 + 1=',
            [Answer(pg.oneof([1, 2, 3]))], lm=fake.Echo()
        ),
        [0.0, -1.0, -2.0]
    )

  def test_score_returning_scoring_results(self):
    self.assertEqual(scoring.score(
        'hi', [1, 2], lm=fake.Echo(), return_scoring_results=True),
                     [lf.LMScoringResult(score=0.0, gradients=None),
                      lf.LMScoringResult(score=-1.0, gradients=None)])

  def test_scope_with_lm_from_the_context(self):
    with lf.context(lm=fake.Echo()):
      self.assertEqual(scoring.score('hi', [1, 2]), [0.0, -1.0])


if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for language model."""

import copy
import os
import tempfile
import time
import unittest

import langfun.core as lf
from langfun.core.llms import fake
from langfun.core.llms.cache import base
from langfun.core.llms.cache import in_memory

import pyglove as pg


class InMemoryLMCacheTest(unittest.TestCase):

  def test_basics(self):
    cache = in_memory.InMemory()
    lm = fake.StaticSequence(['1', '2', '3', '4', '5', '6'], cache=cache)
    self.assertEqual(lm('a'), '1')
    self.assertEqual(lm('a', cache_seed=1), '2')
    self.assertEqual(lm('b'), '3')
    self.assertEqual(lm('a'), '1')
    self.assertEqual(lm('c'), '4')
    self.assertEqual(lm('a', cache_seed=None), '5')
    self.assertEqual(lm('a', cache_seed=None), '6')

    self.assertEqual(cache.model_ids(), ['StaticSequence'])
    self.assertEqual(
        list(cache.keys()),
        [
            ('a', (None, None, 1, 40, None, None), 0),
            ('a', (None, None, 1, 40, None, None), 1),
            ('b', (None, None, 1, 40, None, None), 0),
            ('c', (None, None, 1, 40, None, None), 0),
        ],
    )
    self.assertEqual(
        list(cache.keys('StaticSequence')),
        [
            ('a', (None, None, 1, 40, None, None), 0),
            ('a', (None, None, 1, 40, None, None), 1),
            ('b', (None, None, 1, 40, None, None), 0),
            ('c', (None, None, 1, 40, None, None), 0),
        ],
    )

    def cache_entry(response_text, cache_seed=0):
      return base.LMCacheEntry(
          lf.LMSamplingResult(
              [
                  lf.LMSample(
                      lf.AIMessage(response_text, cache_seed=cache_seed),
                      score=1.0,
                  )
              ],
              usage=lf.LMSamplingUsage(
                  1,
                  len(response_text),
                  len(response_text) + 1,
              ),
              is_cached=True,
          )
      )

    self.assertEqual(
        list(cache.values()),
        [
            cache_entry('1'),
            cache_entry('2', 1),
            cache_entry('3'),
            cache_entry('4'),
        ],
    )
    self.assertEqual(
        list(cache.values('StaticSequence')),
        [
            cache_entry('1'),
            cache_entry('2', 1),
            cache_entry('3'),
            cache_entry('4'),
        ],
    )
    self.assertEqual(
        list(cache.items()),
        [
            (
                ('a', (None, None, 1, 40, None, None), 0),
                cache_entry('1'),
            ),
            (
                ('a', (None, None, 1, 40, None, None), 1),
                cache_entry('2', 1),
            ),
            (
                ('b', (None, None, 1, 40, None, None), 0),
                cache_entry('3'),
            ),
            (
                ('c', (None, None, 1, 40, None, None), 0),
                cache_entry('4'),
            ),
        ],
    )
    self.assertEqual(
        list(cache.items('StaticSequence')),
        [
            (
                ('a', (None, None, 1, 40, None, None), 0),
                cache_entry('1'),
            ),
            (
                ('a', (None, None, 1, 40, None, None), 1),
                cache_entry('2', 1),
            ),
            (
                ('b', (None, None, 1, 40, None, None), 0),
                cache_entry('3'),
            ),
            (
                ('c', (None, None, 1, 40, None, None), 0),
                cache_entry('4'),
            ),
        ],
    )

    # Test clone/copy semantics.
    self.assertIs(cache.clone()._stats, cache._stats)
    self.assertIs(cache.clone()._cache, cache._cache)
    self.assertIs(cache.clone(deep=True)._cache, cache._cache)
    self.assertIs(cache.clone(deep=True)._stats, cache._stats)
    self.assertIs(copy.copy(cache)._cache, cache._cache)
    self.assertIs(copy.copy(cache)._stats, cache._stats)
    self.assertIs(copy.deepcopy(cache)._cache, cache._cache)
    self.assertIs(copy.deepcopy(cache)._stats, cache._stats)

    self.assertFalse(
        cache.delete(fake.StaticResponse('hi'), lf.UserMessage('c'), seed=0)
    )
    self.assertFalse(cache.delete(lm, lf.UserMessage('c'), seed=1))
    self.assertFalse(cache.delete(lm, lf.UserMessage('d'), seed=0))
    self.assertTrue(cache.delete(lm, lf.UserMessage('c'), seed=0))
    self.assertEqual(
        list(cache.keys('StaticSequence')),
        [
            ('a', (None, None, 1, 40, None, None), 0),
            ('a', (None, None, 1, 40, None, None), 1),
            ('b', (None, None, 1, 40, None, None), 0),
        ],
    )
    self.assertEqual(cache.stats.num_deletes, 1)

  def test_cache_with_modalities(self):

    class CustomModality(lf.Modality):
      content: str

      def to_bytes(self):
        return self.content.encode()

    cache = in_memory.InMemory()
    lm = fake.StaticSequence(['1', '2', '3', '4', '5', '6'], cache=cache)
    lm(lf.UserMessage('hi <<[[image]]>>', image=CustomModality('foo')))
    lm(lf.UserMessage('hi <<[[image]]>>', image=CustomModality('bar')))
    self.assertEqual(
        list(cache.keys()),
        [
            (
                'hi <<[[image]]>><image>acbd18db</image>',
                (None, None, 1, 40, None, None),
                0,
            ),
            (
                'hi <<[[image]]>><image>37b51d19</image>',
                (None, None, 1, 40, None, None),
                0,
            ),
        ],
    )

  def test_ttl(self):
    cache = in_memory.InMemory(ttl=1)
    lm = fake.StaticSequence(['1', '2', '3'], cache=cache)
    self.assertEqual(lm('a'), '1')
    self.assertEqual(lm('a'), '1')
    time.sleep(2)
    self.assertEqual(lm('a'), '2')
    self.assertEqual(cache.stats.num_updates, 2)
    self.assertEqual(cache.stats.num_queries, 3)
    self.assertEqual(cache.stats.num_hits, 1)
    self.assertEqual(cache.stats.num_hit_expires, 1)
    self.assertEqual(cache.stats.num_misses, 1)
    self.assertEqual(cache.stats.num_deletes, 1)

  def test_different_sampling_options(self):
    cache = in_memory.InMemory()
    lm = fake.StaticSequence(['1', '2', '3'], cache=cache)
    self.assertEqual(lm('a'), '1')
    self.assertEqual(lm('a'), '1')
    self.assertEqual(lm('a', temperature=1.0), '2')
    self.assertEqual(
        list(cache.keys()),
        [
            ('a', (None, None, 1, 40, None, None), 0),
            ('a', (1.0, None, 1, 40, None, None), 0),
        ],
    )

  def test_different_model(self):
    cache = in_memory.InMemory()
    lm1 = fake.StaticSequence(['1', '2', '3'], cache=cache, temperature=0.0)
    lm2 = fake.Echo(cache=cache, temperature=0.0)

    self.assertEqual(lm1('a'), '1')
    self.assertEqual(lm2('a'), 'a')
    self.assertEqual(lm1('a'), '1')
    self.assertEqual(lm1('b'), '2')
    self.assertEqual(lm2('b'), 'b')

    self.assertEqual(
        list(cache.keys('StaticSequence')),
        [
            ('a', (0.0, None, 1, 40, None, None), 0),
            ('b', (0.0, None, 1, 40, None, None), 0),
        ],
    )
    self.assertEqual(
        list(cache.keys('Echo')),
        [
            ('a', (0.0, None, 1, 40, None, None), 0),
            ('b', (0.0, None, 1, 40, None, None), 0),
        ],
    )
    self.assertEqual(len(cache), 4)
    cache.reset('Echo')
    self.assertEqual(list(cache.keys('Echo')), [])
    cache.reset()
    self.assertEqual(list(cache.keys()), [])

  def test_save_load(self):
    pg.set_load_handler(pg.symbolic.default_load_handler)
    pg.set_save_handler(pg.symbolic.default_save_handler)

    cache = in_memory.InMemory()
    lm1 = fake.StaticSequence(['1', '2', '3'], cache=cache)
    lm2 = fake.Echo(cache=cache)

    self.assertEqual(lm1('a'), '1')
    self.assertEqual(lm2('a'), 'a')
    self.assertEqual(cache.stats.num_updates, 2)

    tmp_dir = tempfile.gettempdir()
    path = os.path.join(tmp_dir, 'memory.json')

    # Path does not exist at the moment.
    cache1 = in_memory.InMemory(path)
    self.assertEqual(len(cache1._cache), 0)

    # Now save the cache to path.
    with self.assertRaisesRegex(ValueError, '`path` must be specified'):
      cache.save()
    cache.save(path)

    cache2 = in_memory.InMemory(path)
    self.assertEqual(cache2._cache, cache._cache)

    # Do nothing since there is no updates.
    self.assertEqual(cache2.stats.num_updates, 0)
    cache2.save()

    lm1 = fake.StaticSequence(['x', 'y'], cache=cache2)
    lm2 = fake.Echo(cache=cache2)

    self.assertEqual(lm1('a'), '1')
    self.assertEqual(lm2('a'), 'a')

    # A new entry.
    self.assertEqual(lm2('b'), 'b')
    self.assertEqual(lm2('c'), 'c')
    self.assertEqual(cache2.stats.num_updates, 2)
    cache2.save()

    # Corrupted file.
    pg.io.writefile(path, 'bad_content')
    cache3 = in_memory.InMemory(path)
    self.assertEqual(len(cache3), 0)


class LmCacheTest(unittest.TestCase):

  def test_lm_cache(self):
    with in_memory.lm_cache() as c:
      lm = fake.Echo()
      self.assertIs(lm.cache, c)
      lm = fake.Echo(cache=in_memory.InMemory())
      self.assertIsNot(lm.cache, c)

  def test_lm_cache_with_file(self):
    pg.set_load_handler(pg.symbolic.default_load_handler)
    pg.set_save_handler(pg.symbolic.default_save_handler)

    cache = in_memory.InMemory()
    lm = fake.StaticSequence(['1', '2', '3'], cache=cache)
    self.assertEqual(lm('a'), '1')
    self.assertEqual(lm('b'), '2')

    tmp_dir = tempfile.gettempdir()
    path1 = os.path.join(tmp_dir, 'memory1.json')
    cache.save(path1)

    with in_memory.lm_cache(path1) as c1:
      self.assertEqual(len(c1), 2)

      lm = fake.StaticSequence(['4', '5', '6'])
      self.assertEqual(lm('a'), '1')
      self.assertEqual(lm('b'), '2')
      self.assertEqual(lm('c'), '4')

    with in_memory.lm_cache(path1) as c2:
      self.assertEqual(len(c2), 3)


if __name__ == '__main__':
  unittest.main()



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import base64
import unittest
import langfun.core as lf
from langfun.core import modalities as lf_modalities
from langfun.core.data.conversion import gemini  # pylint: disable=unused-import


image_content = (
    b'\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\x18\x00\x00\x00\x18\x04'
    b'\x03\x00\x00\x00\x12Y \xcb\x00\x00\x00\x18PLTE\x00\x00'
    b'\x00fff_chaag_cg_ch^ci_ciC\xedb\x94\x00\x00\x00\x08tRNS'
    b'\x00\n\x9f*\xd4\xff_\xf4\xe4\x8b\xf3a\x00\x00\x00>IDATx'
    b'\x01c \x05\x08)"\xd8\xcc\xae!\x06pNz\x88k\x19\\Q\xa8"\x10'
    b'\xc1\x14\x95\x01%\xc1\n\xa143Ta\xa8"D-\x84\x03QM\x98\xc3'
    b'\x1a\x1a\x1a@5\x0e\x04\xa0q\x88\x05\x00\x07\xf8\x18\xf9'
    b'\xdao\xd0|\x00\x00\x00\x00IEND\xaeB`\x82'
)


class GeminiConversionTest(unittest.TestCase):

  def test_as_format_with_role(self):
    self.assertEqual(
        lf.UserMessage('hi').as_format('gemini'),
        {
            'role': 'user',
            'parts': [{'text': 'hi'}],
        },
    )
    self.assertEqual(
        lf.AIMessage('hi').as_format('gemini'),
        {
            'role': 'model',
            'parts': [{'text': 'hi'}],
        },
    )
    self.assertEqual(
        lf.SystemMessage('hi').as_format('gemini'),
        {
            'role': 'system',
            'parts': [{'text': 'hi'}],
        },
    )

  def test_as_format_with_image(self):
    self.assertEqual(
        lf.Template(
            'What are the common words from {{image}} and {{video}}?',
            image=lf_modalities.Image.from_bytes(image_content),
            video=lf_modalities.Custom.from_uri(
                'https://www.youtube.com/watch?v=abcd', mime='text/html'
            ),
        ).render().as_gemini_format(),
        {
            'role': 'user',
            'parts': [
                {
                    'text': 'What are the common words from'
                },
                {
                    'inlineData': {
                        'data': base64.b64encode(image_content).decode('utf-8'),
                        'mimeType': 'image/png',
                    }
                },
                {
                    'text': 'and'
                },
                {
                    'fileData': {
                        'fileUri': 'https://www.youtube.com/watch?v=abcd',
                        'mimeType': 'video/*',
                    }
                },
                {
                    'text': '?'
                }
            ],
        },
    )

  def test_as_format_with_chunk_preprocessor(self):
    self.assertEqual(
        lf.Template(
            'What is this {{image}}?',
            image=lf_modalities.Image.from_bytes(image_content)
        ).render().as_format(
            'gemini',
            chunk_preprocessor=lambda x: x if isinstance(x, str) else None
        ),
        {
            'role': 'user',
            'parts': [
                {
                    'text': 'What is this'
                },
                {
                    'text': '?'
                }
            ],
        },
    )

  def test_from_value_with_simple_text(self):
    self.assertEqual(
        lf.Message.from_value(
            {
                'parts': [{'text': 'this is a text'}],
            },
            format='gemini',
        ),
        lf.AIMessage('this is a text'),
    )

  def test_from_value_with_role(self):
    self.assertEqual(
        lf.Message.from_value(
            {
                'role': 'user',
                'parts': [{'text': 'this is a text'}],
            },
            format='gemini',
        ),
        lf.UserMessage('this is a text'),
    )
    self.assertEqual(
        lf.Message.from_value(
            {
                'role': 'model',
                'parts': [{'text': 'this is a text'}],
            },
            format='gemini',
        ),
        lf.AIMessage('this is a text'),
    )
    self.assertEqual(
        lf.Message.from_value(
            {
                'role': 'system',
                'parts': [{'text': 'this is a text'}],
            },
            format='gemini',
        ),
        lf.SystemMessage('this is a text'),
    )
    with self.assertRaisesRegex(ValueError, 'Unsupported role: .*'):
      lf.Message.from_value(
          {
              'role': 'function',
              'parts': [{'text': 'this is a text'}],
          },
          format='gemini',
      )

  def test_from_value_with_thoughts(self):
    message = lf.Message.from_value(
        {
            'role': 'user',
            'parts': [
                {
                    'text': 'this is a red round object',
                    'thought': True
                },
                {
                    'text': 'this is a apple',
                },
            ],
        },
        format='gemini',
    )
    self.assertEqual(message.text, 'this is a apple')
    self.assertEqual(message.thought, 'this is a red round object')

  def test_from_value_with_modalities(self):
    m = lf.Message.from_gemini_format(
        {
            'role': 'user',
            'parts': [
                {
                    'text': 'What are the common words from'
                },
                {
                    'inlineData': {
                        'data': base64.b64encode(image_content).decode('utf-8'),
                        'mimeType': 'image/png',
                    }
                },
                {
                    'text': ','
                },
                {
                    'fileData': {
                        'fileUri': 'https://my.pdf',
                        'mimeType': 'application/pdf',
                    }
                },
                {
                    'text': 'and'
                },
                {
                    'fileData': {
                        'fileUri': 'https://www.youtube.com/watch?v=abcd',
                        'mimeType': 'video/*',
                    }
                },
                {
                    'text': '?'
                }
            ],
        },
    )
    self.assertEqual(
        m.text,
        (
            'What are the common words from <<[[obj0]]>> , <<[[obj1]]>> '
            'and <<[[obj2]]>> ?'
        )
    )
    self.assertIsInstance(m.obj0, lf_modalities.Image)
    self.assertEqual(m.obj0.mime_type, 'image/png')
    self.assertEqual(m.obj0.to_bytes(), image_content)

    self.assertIsInstance(m.obj1, lf_modalities.PDF)
    self.assertEqual(m.obj1.uri, 'https://my.pdf')

    self.assertIsInstance(m.obj2, lf_modalities.Video)
    self.assertEqual(m.obj2.uri, 'https://www.youtube.com/watch?v=abcd')


if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import unittest
import langfun.core as lf
from langfun.core.llms import groq


class GroqTest(unittest.TestCase):

  def test_basics(self):
    self.assertEqual(groq.GroqMistral_8x7B().model_id, 'mixtral-8x7b-32768')
    self.assertEqual(
        groq.GroqMistral_8x7B().resource_id, 'groq://mixtral-8x7b-32768'
    )

  def test_request_args(self):
    args = groq.GroqMistral_8x7B()._request_args(
        lf.LMSamplingOptions(
            temperature=1.0, stop=['\n'], n=1, random_seed=123,
            logprobs=True, top_logprobs=True
        )
    )
    self.assertNotIn('logprobs', args)
    self.assertNotIn('top_logprobs', args)

  def test_api_key(self):
    lm = groq.GroqMistral_8x7B()
    with self.assertRaisesRegex(ValueError, 'Please specify `api_key`'):
      _ = lm.headers

    lm = groq.GroqMistral_8x7B(api_key='fake key')
    self.assertEqual(
        lm.headers,
        {
            'Content-Type': 'application/json',
            'Authorization': 'Bearer fake key',
        }
    )

    os.environ['GROQ_API_KEY'] = 'abc'
    lm = groq.GroqMistral_8x7B()
    self.assertEqual(
        lm.headers,
        {
            'Content-Type': 'application/json',
            'Authorization': 'Bearer abc',
        }
    )
    del os.environ['GROQ_API_KEY']

  def test_lm_get(self):
    self.assertIsInstance(
        lf.LanguageModel.get('groq://gemma2-9b-it'),
        groq.Groq,
    )

if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for langfun.core.templates.Completion."""

import unittest
import langfun.core as lf
from langfun.core.llms import fake as lf_llms
from langfun.core.templates.completion import Completion


class CompletionTest(unittest.TestCase):

  def test_prompt_only(self):
    l = Completion(prompt='Yesterday is 2023-06-04, today is ')
    self.assertIsNone(l.response)
    self.assertEqual(l.render(), 'Yesterday is 2023-06-04, today is')

  def test_prompt_and_response(self):
    l = Completion(
        prompt='Yesterday is 2023-06-04, today is ', lm_response='2023-06-05'
    )
    self.assertEqual(
        l.render(), 'Yesterday is 2023-06-04, today is\n2023-06-05')

  def test_call_and_cache(self):
    l = Completion(
        prompt='Yesterday is 2023-06-04, today is ', cache_response=True)

    with lf.context(
        lm=lf_llms.StaticMapping(
            mapping={'Yesterday is 2023-06-04, today is': '2023-06-05'}
        )
    ):
      l()
      self.assertEqual(l.response, '2023-06-05')


if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Structures for Python code generation."""

import functools
from typing import Annotated, Any, Callable, ContextManager
import langfun.core as lf
from langfun.core.coding.python import correction
from langfun.core.coding.python import execution
import pyglove as pg


class PythonCode(pg.Object):
  """Symbolic class for Python code.

  The value of the last expression of the source will be the returned value.
  """

  source: Annotated[
      str,
      'Source code. Avoid using docstring for defined functions and classes.',
  ]

  _TLS_AUTO_RUN = '__auto_run__'

  @classmethod
  def auto_run(
      cls,
      enabled: bool = True,
      sandbox: bool | None = None,
      timeout: int | None = 5) -> ContextManager[None]:
    """Returns a context manager to enable/disable auto run of PythonCode.

    `auto_run` is thread-safe and can be nested. For example::

      with lf.coding.PythonCode.auto_run(True, sandbox=True, timeout=1):
        a = PythonCode('1 + 2')
        assert a == 1
        with lf.coding.PythonCode.auto_run(False):
          b = PythonCode('2 + 3')
          assert isinstance(b, lf.PythonCode)

    Args:
      enabled: If True, enable auto call for functors.
        Otherwise, auto call will be disabled.
      sandbox: If True, run code in sandbox; If False, run code in current
        process. If None, run in sandbox first, if the output could not be
        serialized and pass to current process, run the code again in current
        process. Applicable when `enabled` is set to True.
      timeout: Timeout in seconds. Applicable when both `enabled` and `sandbox`
        are set to True.

    Returns:
      A context manager for enabling/disabling auto call for functors.
    """
    return pg.object_utils.thread_local_value_scope(
        cls._TLS_AUTO_RUN,
        pg.Dict(enabled=enabled, sandbox=sandbox, timeout=timeout),
        pg.Dict(enabled=False, sandbox=None, timeout=None),
    )

  def __new__(cls, *args, **kwargs):
    instance = object.__new__(cls)
    auto_run = pg.object_utils.thread_local_get(
        cls._TLS_AUTO_RUN,
        pg.Dict(enabled=False, sandbox=None, timeout=None)
    )
    if auto_run.enabled:
      instance.__init__(*args, **kwargs)
      return instance(
          sandbox=auto_run.sandbox, timeout=auto_run.timeout)
    return instance

  def __call__(
      self,
      *,
      sandbox: bool | None = None,
      timeout: int | None = 5,
      global_vars: dict[str, Any] | None = None,
      returns_stdout: bool = False,
      outputs_intermediate: bool = False,
      autofix: int = 3,
      autofix_lm: lf.LanguageModel | None = None,
  ) -> Any:
    """Returns the value of the last expression from the source.

    Args:
      sandbox: If True, run code in sandbox; If False, run code in current
        process. If None, run in sandbox first, if the output could not be
        serialized and pass to current process, run the code again in current
        process.
      timeout: Timeout in seconds. If None, there is no timeout. Applicable when
        sandbox is set to True.
      global_vars: Global variables that could be accessed from the source code.
      returns_stdout: If True, the stdout (a str) will be returned.
      outputs_intermediate: Applicable when returns_stdout is False. If True,
        intermediate output will be outputted as a dict, with the last line's
        value accessible by key '__result__' and the std output accessible by 
        key '__stdout__'. Otherwise the value of the last line will be returned.
      autofix: Number of attempts to auto fix the generated code. If 0, autofix
        is disabled.
      autofix_lm: Language model to be used. If not specified, it will try to
        use the `lm` under `lf.context`.

    Returns:
      The value of the last expression in the source code. Or a dict of local
      variable names defined in the source code to their values if
      `outputs_intermediate` is set to True. The value for the last line can be
      accessed by key '__result__'. Or the stdout as a str if `returns_stdout`
      is set to True.

    Raises:
      TimeoutError: If `sandbox` is True and timeout has reached.
      Exception: Any errors that the source code has raised.
    """
    result, updated_code = correction.run_with_correction(
        self.source,
        global_vars=global_vars,
        sandbox=sandbox,
        timeout=timeout,
        max_attempts=autofix,
        lm=autofix_lm,
        returns_code=True,
        returns_stdout=returns_stdout,
        outputs_intermediate=outputs_intermediate,
    )
    self.rebind(source=updated_code)
    return result

  def eval(
      self,
      *,
      sandbox: bool | None = None,
      timeout: int | None = 5,
      global_vars: dict[str, Any] | None = None,
      autofix: int = 3,
      autofix_lm: lf.LanguageModel = lf.contextual(),
  ) -> Any | tuple[Any, str]:
    """Evaluates the code and return a dict of local variable names to values.

    Args:
      sandbox: If True, run code in sandbox; If False, run code in current
        process. If None, run in sandbox first, if the output could not be
        serialized and pass to current process, run the code again in current
        process.
      timeout: Timeout in seconds. If None, there is no timeout. Applicable when
        sandbox is set to True.
      global_vars: Global variables that could be accessed from the source code.
      autofix: Number of attempts to auto fix the generated code. If 0, autofix
        is disabled. Auto-fix is not supported for 'json' protocol.
      autofix_lm: Language model to be used. If not specified, it will try to
        use the `lm` under `lf.context`.

    Returns:
      A dict of local variable names defined in the source code to their values
      if `returns_code` is set to False (default), otherwise a tuple
      of (dict, final code str).

    Raises:
      TimeoutError: If `sandbox` is True and timeout has reached.
      Exception: Any errors that the source code has raised.
    """
    return self(
        sandbox=sandbox,
        timeout=timeout,
        global_vars=global_vars,
        autofix=autofix,
        autofix_lm=autofix_lm,
        outputs_intermediate=True,
    )


class PythonFunction(pg.Object):
  """Generated Python function via source code.

  The source code will be directly passed into eval() for execution and the
  output of the function will be returned.
  """

  name: str
  args: dict[str, str]
  returns: str
  source: Annotated[str, 'Source code for the Python function. ']

  def _on_bound(self):
    super()._on_bound()
    self.__dict__.pop('implementation', None)

  @functools.cached_property
  def implementation(self) -> Callable[..., Any]:
    """Returns the function implementation based on source code."""
    return execution.run(self.source)

  def __call__(
      self,
      *args,
      sandbox: bool | None = None,
      timeout: int | None = 5,
      **kwargs) -> Any:
    """Calls the function.

    Args:
      *args: Positional arguments that will be passed to the implementation.
      sandbox: If True, run code in sandbox; If False, run code in current
        process. If None, run in sandbox first, if the output could not be
        serialized and pass to current process, run the code again in current
        process.
      timeout: Timeout in seconds. If None, there is no timeout. Applicable when
        sandbox is set to True.
      **kwargs: Keyword arguments that will be passed to the implementation.

    Returns:
      The return value of the implementation.

    Raises:
      TimeoutError: If `sandbox` is True and timeout has reached.
      Exception: Any errors that the source code has raised.
    """
    return pg.coding.maybe_sandbox_call(
        self.implementation, *args, sandbox=sandbox, timeout=timeout, **kwargs
    )



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import inspect
import unittest

from langfun.core.llms import fake
from langfun.core.structured import mapping
from langfun.core.structured import schema_generation


class GenerateClassTest(unittest.TestCase):

  def test_generate_class_prompt(self):
    input_message = schema_generation.generate_class(
        'Trip',
        'Generate a trip class',
        skip_lm=True,
        returns_message=True,
    )
    self.maxDiff = None
    self.assertEqual(
        input_message.text,
        inspect.cleandoc("""
            Help generate a class based on the last CLASS_NAME and GENERATION_CONTEXT.

            Instructions:
            - Use `Object` as the base class for all generated classes
            - Create auxillary classes for composition if needed.
            - Use Python type annotation for declaraing fields:
              (e.g. bool, str, int, float, Optional[str], List[int], Union[str, int])
            - Do not use types that need import.
            - Avoid self-referential types. e.g:
              ```
              class Node(Object):
                children: list[Node]
              ```
            - Do not generate methods.

            CLASS_NAME:
              Solution

            GENERATION_CONTEXT:
              How to evaluate an arithmetic expression?

            OUTPUT_CLASS:
              ```python
              class Step(Object):
                description: str
                output: float

              class Solution(Object):
                steps: list[Step]
                result: float
              ```


            CLASS_NAME:
              Trip

            GENERATION_CONTEXT:
              Generate a trip class

            OUTPUT_CLASS:
            """),
    )

  def test_generate_class(self):
    lm = fake.StaticResponse("""
        ```python
        class A(Object):
          x: int

        class B(Object):
          a: A
        ```
        """)
    cls = schema_generation.generate_class(
        'B',
        'Generate a B class with a field pointing to another class A',
        lm=lm,
    )
    self.assertIs(cls.__name__, 'B')

    with self.assertRaises(mapping.MappingError):
      schema_generation.generate_class(
          'Foo',
          'Generate a Foo class with a field pointing to another class A',
          lm=lm,
      )


if __name__ == '__main__':
  unittest.main()



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for action evaluation."""

import os
import tempfile
import unittest

from langfun.core import eval as lf_eval
from langfun.core import llms as lf_llms
from langfun.core.agentic import action as action_lib
from langfun.core.agentic import action_eval
import pyglove as pg


class Foo(action_lib.Action):
  x: int

  def call(self, session, **kwargs):
    del session, kwargs
    return self.x


@pg.functor()
def foo_inputs():
  return [
      pg.Dict(action=Foo(1), groundtruth=1),
      pg.Dict(action=Foo(2), groundtruth=1),
  ]


class ActionEvalTest(unittest.TestCase):

  def test_basics(self):

    class FooEval(action_eval.ActionEval):
      inputs = foo_inputs()
      metrics = [lf_eval.v2.metrics.Match()]
      action_args = dict(
          lm=lf_llms.Echo()
      )

    s = FooEval()
    root_dir = os.path.join(tempfile.gettempdir(), 'foo_eval')
    s.run(root_dir, plugins=[])
    self.assertEqual(s.metrics[0].matches, 0.5)
    self.assertEqual(s.metrics[0].mismatches, 0.5)


class ActionEvalV1Test(unittest.TestCase):

  def test_basics(self):

    class FooEval(action_eval.ActionEvalV1):
      lm = lf_llms.Echo()
      inputs = foo_inputs()

    s = FooEval()
    result = s.run(summary=False)
    self.assertEqual(
        result,
        dict(
            experiment_setup=dict(
                id=s.id,
                dir=None,
                model='Echo',
                prompt_template='<unused>',
                method='query',
                schema_fn='_dummy_schema()'
            ),
            cache_stats=dict(
                use_cache=True,
                num_queries=0,
                num_hits=0,
                num_updates=0,
            ),
            metrics=dict(
                total=2,
                failures=0,
                failure_rate=0.0,
                oop_failures=0,
                oop_failure_rate=0.0,
                non_oop_failures=0,
                non_oop_failure_rate=0.0,
                failure_breakdown={},
                num_matches=0,
                match_rate=0.0,
                num_mismatches=2,
                mismatch_rate=1.0
            ),
            usage=None
        )
    )


if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Schema for structured data."""

import abc
import inspect
import io
import re
import sys
import textwrap
import typing
from typing import Any, Literal, Sequence, Type, Union
import langfun.core as lf
from langfun.core.coding.python import correction
import pyglove as pg


def include_method_in_prompt(method):
  """Decorator to include a method in the class definition of the prompt."""
  setattr(method, '__show_in_prompt__', True)
  return method


def should_include_method_in_prompt(method):
  """Returns true if the method should be shown in the prompt."""
  return getattr(method, '__show_in_prompt__', False)


def parse_value_spec(value) -> pg.typing.ValueSpec:
  """Parses a PyGlove ValueSpec equivalence into a ValueSpec."""
  if isinstance(value, pg.typing.ValueSpec):
    return value

  if isinstance(value, dict) and len(value) == 1 and 'result' in value:
    value = value['result']

  def _parse_node(v) -> pg.typing.ValueSpec:
    if isinstance(v, dict):
      return pg.typing.Dict([(k, _parse_node(cv)) for k, cv in v.items()])
    elif isinstance(v, list):
      if len(v) != 1:
        raise ValueError(
            'Annotation with list must be a list of a single element. '
            f'Encountered: {v}'
        )
      return pg.typing.List(_parse_node(v[0]))
    else:
      spec = pg.typing.ValueSpec.from_annotation(v, auto_typing=True)
      if isinstance(
          spec,
          (
              pg.typing.Any,
              pg.typing.Callable,
              pg.typing.Tuple,
              pg.typing.Type,
              pg.typing.Union,
          ),
      ):
        raise ValueError(f'Unsupported schema specification: {v}')
      return spec

  return _parse_node(value)


SchemaProtocol = Literal['json', 'python']


class SchemaError(Exception):   # pylint: disable=g-bad-exception-name
  """Schema error."""

  def __init__(self,
               schema: 'Schema',
               value: Any,
               protocol: SchemaProtocol,
               cause: Exception):
    self.schema = schema
    self.value = value
    self.protocol = protocol
    self.cause = cause

  def __str__(self):
    r = io.StringIO()
    r.write(
        pg.colored(
            f'{self.cause.__class__.__name__}: {self.cause}', 'magenta'
        )
    )

    r.write('\n')
    r.write(pg.colored('Schema:', 'red'))
    r.write('\n\n')
    r.write(textwrap.indent(
        pg.colored(
            schema_repr(self.protocol).repr(self.schema), 'magenta'
        ),
        ' ' * 2
    ))
    r.write('\n\n')
    r.write(pg.colored('Generated value:', 'red'))
    r.write('\n\n')
    r.write(textwrap.indent(
        pg.colored(value_repr(self.protocol).repr(self.value), 'magenta'),
        ' ' * 2
    ))
    return r.getvalue()


class Schema(
    lf.NaturalLanguageFormattable,
    pg.Object,
    pg.views.HtmlTreeView.Extension
):
  """Base class for structured data schema."""

  spec: pg.typing.Annotated[
      pg.typing.Object(pg.typing.ValueSpec, transform=parse_value_spec),
      (
          'A PyGlove ValueSpec object representing the spec for the value '
          'to be parsed.'
      ),
  ]

  def schema_str(self, protocol: SchemaProtocol = 'json', **kwargs) -> str:
    """Returns the representation of the schema."""
    return schema_repr(protocol).repr(self, **kwargs)

  def value_str(
      self, value: Any, protocol: SchemaProtocol = 'json', **kwargs
  ) -> str:
    """Returns the representation of a structured value."""
    return value_repr(protocol).repr(value, self, **kwargs)

  def parse(
      self, text: str, protocol: SchemaProtocol = 'json', **kwargs
  ) -> Any:
    """Parse a LM generated text into a structured value."""
    value = value_repr(protocol).parse(text, self, **kwargs)

    # TODO(daiyip): support autofix for schema error.
    try:
      return self.spec.apply(value)
    except Exception as e:
      raise SchemaError(self, value, protocol, e)  # pylint: disable=raise-missing-from

  def natural_language_format(self) -> str:
    return self.schema_str()

  def schema_dict(self) -> dict[str, Any]:
    """Returns the dict representation of the schema."""

    def _node(vs: pg.typing.ValueSpec) -> Any:
      if isinstance(vs, pg.typing.PrimitiveType):
        return vs
      elif isinstance(vs, pg.typing.Dict):
        assert vs.schema is not None
        return {str(k): _node(f.value) for k, f in vs.schema.fields.items()}
      elif isinstance(vs, pg.typing.List):
        return [_node(vs.element.value)]
      elif isinstance(vs, pg.typing.Object):
        if issubclass(vs.cls, pg.Object):
          d = {pg.JSONConvertible.TYPE_NAME_KEY: vs.cls.__serialization_key__}
          d.update(
              {
                  str(k): _node(f.value)
                  for k, f in vs.cls.__schema__.fields.items()
              }
          )
          return d
      raise TypeError(
          'Unsupported value spec being used as the schema for '
          f'structured data: {vs}.')

    return {'result': _node(self.spec)}

  def class_dependencies(
      self,
      include_base_classes: bool = True,
      include_subclasses: bool = True,
      include_generated_subclasses: bool = False) -> list[Type[Any]]:
    """Returns a list of class dependencies for current schema."""
    return class_dependencies(
        self.spec,
        include_base_classes,
        include_subclasses,
        include_generated_subclasses
    )

  @classmethod
  def from_value(cls, value) -> 'Schema':
    """Creates a schema from an equivalent representation."""
    if isinstance(value, Schema):
      return value
    return cls(parse_value_spec(value))

  def _html_tree_view_content(
      self,
      *,
      view: pg.views.HtmlTreeView,
      **kwargs,
  ):
    return pg.Html.element(
        'div',
        [pg.Html.escape(self.schema_str(protocol='python'))],
        css_classes=['lf-schema-definition']
    ).add_style(
        """
        .lf-schema-definition {
            color: blue;
            margin: 5px;
            white-space: pre-wrap;
        }
        """
    )


SchemaType = Union[Schema, Type[Any], list[Type[Any]], dict[str, Any]]


def _top_level_object_specs_from_value(value: pg.Symbolic) -> list[Type[Any]]:
  """Returns a list of top level value specs from a symbolic value."""
  top_level_object_specs = []

  def _collect_top_level_object_specs(k, v, p):
    del k, p
    if isinstance(v, pg.Object):
      top_level_object_specs.append(pg.typing.Object(v.__class__))
      return pg.TraverseAction.CONTINUE
    return pg.TraverseAction.ENTER

  pg.traverse(value, _collect_top_level_object_specs)
  return top_level_object_specs


def class_dependencies(
    value_or_spec: Union[
        pg.Symbolic,
        Schema,
        pg.typing.ValueSpec,
        Type[pg.Object],
        tuple[Union[pg.typing.ValueSpec, Type[pg.Object]], ...],
    ],
    include_base_classes: bool = True,
    include_subclasses: bool = True,
    include_generated_subclasses: bool = False,
) -> list[Type[Any]]:
  """Returns a list of class dependencies from a value or specs."""
  if isinstance(value_or_spec, Schema):
    value_or_spec = value_or_spec.spec

  if inspect.isclass(value_or_spec) or isinstance(
      value_or_spec, pg.typing.ValueSpec
  ):
    value_or_spec = (value_or_spec,)

  if isinstance(value_or_spec, tuple):
    value_specs = []
    for v in value_or_spec:
      if isinstance(v, pg.typing.ValueSpec):
        value_specs.append(v)
      elif inspect.isclass(v):
        value_specs.append(pg.typing.Object(v))
      else:
        raise TypeError(f'Unsupported spec type: {v!r}')
  else:
    value_specs = _top_level_object_specs_from_value(value_or_spec)

  seen = set()
  dependencies = []

  def _add_dependency(cls_or_classes):
    if isinstance(cls_or_classes, type):
      cls_or_classes = [cls_or_classes]
    for cls in cls_or_classes:
      if cls not in dependencies:
        dependencies.append(cls)

  def _fill_dependencies(vs: pg.typing.ValueSpec, include_subclasses: bool):
    if isinstance(vs, pg.typing.Object):
      if vs.cls not in seen:
        seen.add(vs.cls)

        if include_base_classes:
          # Add base classes as dependencies.
          for base_cls in vs.cls.__bases__:
            # We only keep track of user-defined symbolic classes.
            if base_cls is not object and base_cls is not pg.Object:
              _fill_dependencies(
                  pg.typing.Object(base_cls), include_subclasses=False
              )

        # Add members as dependencies.
        for field in pg.schema(vs.cls).values():
          _fill_dependencies(field.value, include_subclasses)
      _add_dependency(vs.cls)

      # Check subclasses if available.
      if include_subclasses:
        for cls in vs.cls.__subclasses__():
          # NOTE(daiyip): To prevent LLM-generated "hallucinated" classes from
          # polluting the generation space, classes dynamically created by
          # 'eval' (which have __module__ == 'builtins') are excluded from
          # dependencies by default.
          if ((include_generated_subclasses or cls.__module__ != 'builtins')
              and cls not in dependencies):
            _fill_dependencies(pg.typing.Object(cls), include_subclasses=True)

    if isinstance(vs, pg.typing.List):
      _fill_dependencies(vs.element.value, include_subclasses)
    elif isinstance(vs, pg.typing.Tuple):
      for elem in vs.elements:
        _fill_dependencies(elem.value, include_subclasses)
    elif isinstance(vs, pg.typing.Dict) and vs.schema:
      for v in vs.schema.values():
        _fill_dependencies(v.value, include_subclasses)
    elif isinstance(vs, pg.typing.Union):
      for v in vs.candidates:
        _fill_dependencies(v, include_subclasses)

  for value_spec in value_specs:
    _fill_dependencies(value_spec, include_subclasses)
  return dependencies


def schema_spec(noneable: bool = False) -> pg.typing.ValueSpec:  # pylint: disable=unused-argument
  if typing.TYPE_CHECKING:
    return Any
  return pg.typing.Object(
      Schema, transform=Schema.from_value, is_noneable=noneable
  )  # pylint: disable=unreachable-code


#
# Schema representations.
#


class SchemaRepr(metaclass=abc.ABCMeta):
  """Base class for schema representation."""

  @abc.abstractmethod
  def repr(self, schema: Schema) -> str:
    """Returns the representation of the schema."""


class SchemaPythonRepr(SchemaRepr):
  """Python-representation for a schema."""

  def repr(
      self,
      schema: Schema,
      *,
      include_result_definition: bool = True,
      markdown: bool = True,
      **kwargs,
  ) -> str:
    ret = ''
    if include_result_definition:
      ret += self.result_definition(schema)
    class_definition_str = self.class_definitions(
        schema, markdown=markdown, **kwargs
    )
    if class_definition_str:
      ret += f'\n\n{class_definition_str}'
    return ret.strip()

  def class_definitions(
      self,
      schema: Schema,
      additional_dependencies: list[Type[Any]] | None = None,
      **kwargs
  ) -> str | None:
    """Returns a string containing of class definitions from a schema."""
    deps = schema.class_dependencies(
        include_base_classes=False, include_subclasses=True
    )
    allowed_dependencies = set(deps)
    if additional_dependencies:
      allowed_dependencies.update(additional_dependencies)
    return class_definitions(
        deps, allowed_dependencies=allowed_dependencies, **kwargs)

  def result_definition(self, schema: Schema) -> str:
    return annotation(schema.spec)


def source_form(value, compact: bool = True, markdown: bool = False) -> str:
  """Returns the source code form of an object."""
  return ValuePythonRepr().repr(value, compact=compact, markdown=markdown)


def class_definitions(
    classes: Sequence[Type[Any]],
    *,
    allowed_dependencies: set[Type[Any]] | None = None,
    strict: bool = False,
    markdown: bool = False,
) -> str | None:
  """Returns a str for class definitions."""
  if not classes:
    return None
  def_str = io.StringIO()
  for i, cls in enumerate(classes):
    if i > 0:
      def_str.write('\n')
    def_str.write(
        class_definition(
            cls,
            strict=strict,
            allowed_dependencies=allowed_dependencies,
        )
    )
  ret = def_str.getvalue()
  if markdown and ret:
    ret = f'```python\n{ret}```'
  return ret


def class_definition(
    cls,
    strict: bool = False,
    allowed_dependencies: set[Type[Any]] | None = None,
) -> str:
  """Returns the Python class definition."""
  out = io.StringIO()
  schema = pg.schema(cls)
  eligible_bases = []
  for base_cls in cls.__bases__:
    if base_cls is not object:
      if allowed_dependencies is None or base_cls in allowed_dependencies:
        eligible_bases.append(base_cls.__name__)

  if eligible_bases:
    base_cls_str = ', '.join(eligible_bases)
    out.write(f'class {cls.__name__}({base_cls_str}):\n')
  else:
    out.write(f'class {cls.__name__}:\n')

  if cls.__doc__:
    doc_lines = cls.__doc__.strip().split('\n')
    if len(doc_lines) == 1:
      out.write(f'  """{cls.__doc__}"""\n')
    else:
      out.write('  """')

      # Since Python 3.13, the indentation of docstring lines is removed.
      # Therefore, we add two spaces to each non-empty line to keep the
      # indentation consistent with the class definition.
      if sys.version_info >= (3, 13):
        for i in range(1, len(doc_lines)):
          if doc_lines[i]:
            doc_lines[i] = ' ' * 2 + doc_lines[i]

      for line in doc_lines:
        out.write(line)
        out.write('\n')
      out.write('  """\n')

  empty_class = True
  if schema.fields:
    for key, field in schema.items():
      if not isinstance(key, pg.typing.ConstStrKey):
        pg.logging.warning(
            'Variable-length keyword arguments is not supported in '
            f'structured parsing or query. Encountered: {cls}, Schema: {schema}'
        )
        continue

      # Skip fields that are marked as excluded from the prompt sent to LLM
      # for OOP.
      if field.metadata.get('exclude_from_prompt', False):
        continue

      # Write field doc string as comments before the field definition.
      if field.description:
        for line in field.description.split('\n'):
          if line:
            out.write('  # ')
            out.write(line)
            out.write('\n')

      annotation_str = annotation(
          field.value, strict=strict, allowed_dependencies=allowed_dependencies
      )
      out.write(f'  {field.key}: {annotation_str}')
      out.write('\n')
      empty_class = False

  for method in _iter_newly_defined_methods(cls, allowed_dependencies):
    source = inspect.getsource(method)
    # Remove decorators from the method definition.
    source = re.sub(r'\s*@.*\.include_method_in_prompt.*\n', '', source)
    out.write('\n')
    out.write(
        textwrap.indent(
            inspect.cleandoc('\n' + source), ' ' * 2)
    )
    out.write('\n')
    empty_class = False

  if empty_class:
    out.write('  pass\n')
  return out.getvalue()


def _iter_newly_defined_methods(
    cls, allowed_dependencies: set[Type[Any]] | None):
  names = {attr_name: True for attr_name in dir(cls)}
  for base in cls.__bases__:
    if allowed_dependencies is None or base in allowed_dependencies:
      for name in dir(base):
        names.pop(name, None)
  for name in names.keys():
    attr = getattr(cls, name)
    if callable(attr) and should_include_method_in_prompt(attr):
      yield attr


def annotation(
    vs: pg.typing.ValueSpec,
    annotate_optional: bool = True,
    strict: bool = False,
    allowed_dependencies: set[Type[Any]] | None = None,
) -> str:
  """Returns the annotation string for a value spec."""
  child_annotation_kwargs = dict(
      strict=strict, allowed_dependencies=allowed_dependencies
  )
  if isinstance(vs, pg.typing.Any):
    return 'Any'
  elif isinstance(vs, pg.typing.Enum):
    candidate_str = ', '.join([repr(v) for v in vs.values])
    return f'Literal[{candidate_str}]'
  elif isinstance(vs, pg.typing.Union):
    candidate_str = ', '.join(
        [
            annotation(c, annotate_optional=False, **child_annotation_kwargs)
            for c in vs.candidates
        ]
    )
    if vs.is_noneable:
      candidate_str += ', None'
    return f'Union[{candidate_str}]'

  if isinstance(vs, pg.typing.Bool):
    x = 'bool'
  elif isinstance(vs, pg.typing.Str):
    if vs.regex is None:
      x = 'str'
    else:
      if strict:
        x = f"pg.typing.Str(regex='{vs.regex.pattern}')"
      else:
        x = f"str(regex='{vs.regex.pattern}')"
  elif isinstance(vs, pg.typing.Number):
    constraints = []
    min_label = 'min_value' if strict else 'min'
    max_label = 'max_value' if strict else 'max'
    if vs.min_value is not None:
      constraints.append(f'{min_label}={vs.min_value}')
    if vs.max_value is not None:
      constraints.append(f'{max_label}={vs.max_value}')
    x = 'int' if isinstance(vs, pg.typing.Int) else 'float'
    if constraints:
      if strict:
        x = (
            'pg.typing.Int'
            if isinstance(vs, pg.typing.Int)
            else 'pg.typing.Float'
        )
      x += '(' + ', '.join(constraints) + ')'
  elif isinstance(vs, pg.typing.Object):
    if allowed_dependencies is None or vs.cls in allowed_dependencies:
      x = vs.cls.__name__
    else:
      x = 'Any'
  elif isinstance(vs, pg.typing.List):
    item_str = annotation(vs.element.value, **child_annotation_kwargs)
    x = f'list[{item_str}]'
  elif isinstance(vs, pg.typing.Tuple):
    elem_str = ', '.join(
        [annotation(el.value, **child_annotation_kwargs) for el in vs.elements]
    )
    x = f'tuple[{elem_str}]'
  elif isinstance(vs, pg.typing.Dict):
    kv_pairs = None
    if vs.schema is not None:
      kv_pairs = [
          (k, annotation(f.value, **child_annotation_kwargs))
          for k, f in vs.schema.items()
          if isinstance(k, pg.typing.ConstStrKey)
      ]

    if kv_pairs:
      kv_str = ', '.join(f"'{k}': {v}" for k, v in kv_pairs)
      x = '{' + kv_str + '}'
      if strict:
        x = f'pg.typing.Dict({x})'
    elif vs.schema and vs.schema.dynamic_field:
      v = annotation(vs.schema.dynamic_field.value, **child_annotation_kwargs)
      x = f'dict[str, {v}]'
    else:
      x = 'dict[str, Any]'

  else:
    raise TypeError(f'Unsupported value spec being used as schema: {vs}.')

  if annotate_optional and vs.is_noneable:
    x += ' | None'
  return x


class SchemaJsonRepr(SchemaRepr):
  """JSON-representation for a schema."""

  def repr(self, schema: Schema, **kwargs) -> str:
    del kwargs
    out = io.StringIO()
    def _visit(node: Any) -> None:
      if isinstance(node, str):
        out.write(f'"{node}"')
      elif isinstance(node, list):
        assert len(node) == 1, node
        out.write('[')
        _visit(node[0])
        out.write(']')
      elif isinstance(node, dict):
        out.write('{')
        for i, (k, v) in enumerate(node.items()):
          if i != 0:
            out.write(', ')
          out.write(f'"{k}": ')
          _visit(v)
        out.write('}')
      elif isinstance(node, pg.typing.Enum):
        out.write(' | '.join(
            f'"{v}"' if isinstance(v, str) else repr(v)
            for v in node.values))
      elif isinstance(node, pg.typing.PrimitiveType):
        x = node.value_type.__name__
        if isinstance(node, pg.typing.Number):
          params = []
          if node.min_value is not None:
            params.append(f'min={node.min_value}')
          if node.max_value is not None:
            params.append(f'max={node.max_value}')
          if params:
            x += f'({", ".join(params)})'
        elif isinstance(node, pg.typing.Str):
          if node.regex is not None:
            x += f'(regex={node.regex.pattern})'
        if node.is_noneable:
          x = x + ' | None'
        out.write(x)
      else:
        raise ValueError(
            f'Unsupported value spec being used as schema: {node}.')
    _visit(schema.schema_dict())
    return out.getvalue()


#
# Value representations.
#


class ValueRepr(metaclass=abc.ABCMeta):
  """Base class for value representation."""

  @abc.abstractmethod
  def repr(self, value: Any, schema: Schema | None = None, **kwargs) -> str:
    """Returns the representation of a structured value."""

  @abc.abstractmethod
  def parse(self, text: str, schema: Schema | None = None, **kwargs) -> Any:
    """Parse a LM generated text into a structured value."""


class ValuePythonRepr(ValueRepr):
  """Python-representation for value."""

  def repr(self,
           value: Any,
           schema: Schema | None = None,
           *,
           compact: bool = True,
           verbose: bool = False,
           markdown: bool = True,
           assign_to_var: str | None = None,
           **kwargs) -> str:
    del schema
    if inspect.isclass(value):
      cls_schema = Schema.from_value(value)
      if isinstance(cls_schema.spec, pg.typing.Object):
        object_code = SchemaPythonRepr().class_definitions(
            cls_schema,
            markdown=markdown,
            # We add `pg.Object` as additional dependencies to the class
            # definition so exemplars for class generation could show
            # pg.Object as their bases.
            additional_dependencies=[pg.Object]
        )
        assert object_code is not None
        return object_code
      else:
        object_code = SchemaPythonRepr().result_definition(cls_schema)
    elif isinstance(value, lf.Template):
      return str(value)
    else:
      object_code = pg.format(
          value, compact=compact, verbose=verbose, python_format=True
      )
      if assign_to_var is not None:
        object_code = f'{assign_to_var} = {object_code}'
    if markdown:
      return f'```python\n{ object_code }\n```'
    return object_code

  def parse(
      self,
      text: str,
      schema: Schema | None = None,
      *,
      additional_context: dict[str, Type[Any]] | None = None,
      permission: pg.coding.CodePermission = (
          pg.coding.CodePermission.ASSIGN | pg.coding.CodePermission.CALL
      ),
      autofix=0,
      autofix_lm: lf.LanguageModel = lf.contextual(),
      **kwargs,
  ) -> Any:
    """Parse a Python string into a structured object."""
    del kwargs
    global_vars = additional_context or {}
    if schema is not None:
      dependencies = schema.class_dependencies()
      global_vars.update({d.__name__: d for d in dependencies})
    return structure_from_python(
        text,
        global_vars=global_vars,
        autofix=autofix,
        autofix_lm=autofix_lm,
        permission=permission,
    )


def structure_from_python(
    code: str,
    *,
    global_vars: dict[str, Any] | None = None,
    permission: pg.coding.CodePermission = (
        pg.coding.CodePermission.ASSIGN | pg.coding.CodePermission.CALL
    ),
    autofix=0,
    autofix_lm: lf.LanguageModel = lf.contextual(),
) -> Any:
  """Evaluates structure from Python code with access to symbols."""
  global_vars = global_vars or {}
  global_vars.update({
      'pg': pg,
      'Object': pg.Object,
      'Any': typing.Any,
      'List': typing.List,
      'Tuple': typing.Tuple,
      'Dict': typing.Dict,
      'Sequence': typing.Sequence,
      'Optional': typing.Optional,
      'Union': typing.Union,
      # Special value markers.
      'UNKNOWN': UNKNOWN,
  })
  # We are creating objects here, so we execute the code without a sandbox.
  return correction.run_with_correction(
      code,
      global_vars=global_vars,
      sandbox=False,
      max_attempts=autofix,
      lm=autofix_lm,
      permission=permission,
  )


class JsonError(Exception):
  """Json parsing error."""

  def __init__(self, json: str, cause: Exception):
    self.json = json
    self.cause = cause

  def __str__(self) -> str:
    r = io.StringIO()
    r.write(
        pg.colored(
            f'{self.cause.__class__.__name__}: {self.cause}', 'magenta'
        )
    )

    r.write('\n\n')
    r.write(pg.colored('JSON text:', 'red'))
    r.write('\n\n')
    r.write(textwrap.indent(pg.colored(self.json, 'magenta'), ' ' * 2))
    return r.getvalue()


class ValueJsonRepr(ValueRepr):
  """JSON-representation for value."""

  def repr(self, value: Any, schema: Schema | None = None, **kwargs) -> str:
    del schema
    return pg.to_json_str(dict(result=value))

  def parse(self, text: str, schema: Schema | None = None, **kwargs) -> Any:
    """Parse a JSON string into a structured object."""
    del schema
    try:
      text = cleanup_json(text)
      v = pg.from_json_str(text, **kwargs)
    except Exception as e:
      raise JsonError(text, e)  # pylint: disable=raise-missing-from

    if not isinstance(v, dict) or 'result' not in v:
      raise JsonError(text, ValueError(
          'The root node of the JSON must be a dict with key `result`. '
          f'Encountered: {v}'
      ))
    return v['result']


def cleanup_json(json_str: str) -> str:
  """Clean up the LM responded JSON string."""
  # Treatments:
  # 1. Extract the JSON string with a top-level dict from the response.
  #    This prevents the leading and trailing texts in the response to
  #    be counted as part of the JSON.
  # 2. Escape new lines in JSON values.

  curly_brackets = 0
  under_json = False
  under_str = False
  str_begin = -1

  cleaned = io.StringIO()
  for i, c in enumerate(json_str):
    if c == '{' and not under_str:
      cleaned.write(c)
      curly_brackets += 1
      under_json = True
      continue
    elif not under_json:
      continue

    if c == '}' and not under_str:
      cleaned.write(c)
      curly_brackets -= 1
      if curly_brackets == 0:
        break
    elif c == '"' and json_str[i - 1] != '\\':
      under_str = not under_str
      if under_str:
        str_begin = i
      else:
        assert str_begin > 0
        str_value = json_str[str_begin : i + 1].replace('\n', '\\n')
        cleaned.write(str_value)
        str_begin = -1
    elif not under_str:
      cleaned.write(c)

  if not under_json:
    raise ValueError(f'No JSON dict in the output: {json_str}')

  if curly_brackets > 0:
    raise ValueError(
        f'Malformated JSON: missing {curly_brackets} closing curly braces.'
    )

  return cleaned.getvalue()


def schema_repr(protocol: SchemaProtocol) -> SchemaRepr:
  """Gets a SchemaRepr object from protocol."""
  if protocol == 'json':
    return SchemaJsonRepr()
  elif protocol == 'python':
    return SchemaPythonRepr()
  raise ValueError(f'Unsupported protocol: {protocol}.')


def value_repr(protocol: SchemaProtocol) -> ValueRepr:
  if protocol == 'json':
    return ValueJsonRepr()
  elif protocol == 'python':
    return ValuePythonRepr()
  raise ValueError(f'Unsupported protocol: {protocol}.')


#
# Special value markers.
#


class Missing(pg.Object, pg.typing.CustomTyping):
  """Value marker for a missing field.

  This class differs from pg.MISSING_VALUE in two aspects:
  * When a field is assigned with lf.Missing(), it's considered non-partial.
  * lf.Missing() could format the value spec as Python annotations that are
    consistent with `lf.structured.Schema.schema_repr()`.
  """

  def _on_bound(self):
    super()._on_bound()
    self._value_spec = None

  @property
  def value_spec(self) -> pg.ValueSpec | None:
    """Returns the value spec that applies to the current missing value."""
    return self._value_spec

  def custom_apply(
      self, path: pg.KeyPath, value_spec: pg.ValueSpec, *args, **kwargs
  ) -> tuple[bool, Any]:
    self._value_spec = value_spec
    return (False, self)

  def format(self, *args, **kwargs) -> str:
    if self._value_spec is None:
      return 'MISSING'
    return f'MISSING({annotation(self._value_spec)})'

  @classmethod
  def find_missing(cls, value: Any) -> dict[str, 'Missing']:
    """Lists all missing values contained in the value."""
    missing = {}

    def _visit(k, v, p):
      del p
      if isinstance(v, Missing):
        missing[k] = v
      return pg.TraverseAction.ENTER

    pg.traverse(value, _visit)
    return missing


MISSING = Missing()


def mark_missing(value: Any) -> Any:
  """Replaces pg.MISSING within the value with lf.structured.Missing objects."""
  if isinstance(value, list):
    value = pg.List(value)
  elif isinstance(value, dict):
    value = pg.Dict(value)
  if isinstance(value, pg.Symbolic):

    def _mark_missing(k, v, p):
      del k, p
      if pg.MISSING_VALUE == v:
        v = Missing()
      return v

    return value.rebind(_mark_missing, raise_on_no_change=False)
  return value


class Unknown(pg.Object, pg.typing.CustomTyping):
  """Value marker for a field that LMs could not provide."""

  def custom_apply(self, *args, **kwargs) -> tuple[bool, Any]:
    return (False, self)

  def format(self, *args, **kwargs) -> str:
    return 'UNKNOWN'


UNKNOWN = Unknown()



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import tempfile
import unittest

from langfun.core.eval.v2 import checkpointing
from langfun.core.eval.v2 import eval_test_helper
from langfun.core.eval.v2 import example as example_lib
from langfun.core.eval.v2 import experiment as experiment_lib
from langfun.core.eval.v2 import runners as runners_lib  # pylint: disable=unused-import
import pyglove as pg

Example = example_lib.Example


class SequenceWriterTest(unittest.TestCase):

  def test_basic(self):
    file = os.path.join(tempfile.mkdtemp(), 'test.jsonl')
    writer = checkpointing.SequenceWriter(file)
    example = Example(id=1, input=pg.Dict(x=1), output=2)
    writer.add(example)
    del writer
    self.assertTrue(pg.io.path_exists(file))

  def test_error_handling(self):
    file = os.path.join(tempfile.mkdtemp(), 'test_error_handling.jsonl')
    writer = checkpointing.SequenceWriter(file)
    writer.add(Example(id=1, input=pg.Dict(x=1), output=2))

    def f():
      raise ValueError('Intentional error')

    try:
      writer.add(f())
    except ValueError:
      del writer

    self.assertTrue(pg.io.path_exists(file))
    with pg.io.open_sequence(file, 'r') as f:
      self.assertEqual(len(list(iter(f))), 1)


class ExampleCollector(experiment_lib.Plugin):
  """Collects all examples."""

  def _on_bound(self):
    super()._on_bound()
    self._examples = {}

  @property
  def examples(self) -> dict[int, example_lib.Example]:
    return self._examples

  def on_example_complete(
      self, runner: runners_lib.Runner,
      experiment: experiment_lib.Experiment,
      example: example_lib.Example,
  ):
    assert experiment.is_leaf, None
    self._examples[example.id] = example


class CheckpointerTest(unittest.TestCase):

  def assert_found_in_log(self, experiment, message):
    found_error_log = False
    for log_entry in experiment._log_entries:
      if log_entry.message.startswith(message):
        found_error_log = True
        break
    self.assertTrue(found_error_log)


class PerExampleCheckpointerTest(CheckpointerTest):

  def test_checkpointing(self):
    root_dir = os.path.join(tempfile.mkdtemp(), 'per_example_checkpointer')
    experiment = eval_test_helper.test_experiment()
    checkpoint_filename = 'checkpoint.jsonl'
    checkpointer = checkpointing.PerExampleCheckpointer(checkpoint_filename)
    collector = ExampleCollector()
    run = experiment.run(
        root_dir, 'new', runner='sequential', plugins=[checkpointer, collector]
    )
    num_processed = {}
    for leaf in experiment.leaf_nodes:
      for i in range(leaf.num_examples):
        self.assertIn(i + 1, collector.examples)
        example = collector.examples[i + 1]
        ckpt = run.output_path_for(leaf, f'checkpoint_{example.id}.jsonl')
        self.assertTrue(pg.io.path_exists(ckpt))
        with pg.io.open_sequence(ckpt) as f:
          examples_from_ckpt = list(iter(f))
          # `eval_test_helper.test_experiment` has two TestEvaluation with
          # llm offset 0, which makes that evaluation run twice.
          if example.has_error and leaf.lm.offset == 0:
            self.assertEqual(len(examples_from_ckpt), 2)
          else:
            self.assertEqual(len(examples_from_ckpt), 1)
      if leaf.id not in num_processed:
        self.assertEqual(leaf.progress.num_skipped, 0)
        num_processed[leaf.id] = leaf.progress.num_processed

    # Run again, should skip existing.
    _ = experiment.run(
        root_dir, 'latest', runner='sequential', plugins=[checkpointer]
    )
    for leaf in experiment.leaf_nodes:
      self.assertEqual(leaf.progress.num_skipped, num_processed[leaf.id])

    # Test warm start without reprocess.
    root_dir = os.path.join(tempfile.mkdtemp(), 'per_example_checkpointer2')
    experiment = eval_test_helper.test_experiment()
    _ = experiment.run(
        root_dir, 'new', runner='sequential', plugins=[checkpointer],
        warm_start_from=run.output_root
    )
    for leaf in experiment.leaf_nodes:
      self.assertEqual(leaf.progress.num_skipped, num_processed[leaf.id])

    # Test warm start with reprocess.
    root_dir = os.path.join(tempfile.mkdtemp(), 'per_example_checkpointer3')
    experiment = eval_test_helper.test_experiment()
    _ = experiment.run(
        root_dir, 'new', runner='sequential', plugins=[checkpointer],
        warm_start_from=run.output_root,
        reprocess=True
    )
    for leaf in experiment.leaf_nodes:
      self.assertEqual(leaf.progress.num_skipped, 0)

    root_dir = os.path.join(tempfile.mkdtemp(), 'per_example_checkpointer4')
    experiment = eval_test_helper.test_experiment()
    _ = experiment.run(
        root_dir, 'new', runner='sequential', plugins=[checkpointer],
        warm_start_from=run.output_root,
        reprocess=[1, 2, 3]
    )
    for leaf in experiment.leaf_nodes:
      self.assertEqual(leaf.progress.num_skipped, num_processed[leaf.id] - 3)

  def test_loading_corrupted_checkpoint(self):
    root_dir = os.path.join(
        tempfile.mkdtemp(),
        'per_example_checkpointer_with_corrupted_checkpoint'
    )
    experiment = eval_test_helper.TestEvaluation()
    checkpoint_filename = 'checkpoint.jsonl'
    checkpointer = checkpointing.PerExampleCheckpointer(checkpoint_filename)
    collector = ExampleCollector()

    run = experiment.run(
        root_dir, 'new', runner='sequential', plugins=[checkpointer, collector]
    )
    num_processed = {}
    for i in range(experiment.num_examples):
      self.assertIn(i + 1, collector.examples)
      example = collector.examples[i + 1]
      ckpt = run.output_path_for(experiment, f'checkpoint_{example.id}.jsonl')
      if not example.has_error:
        self.assertTrue(pg.io.path_exists(ckpt))
        with pg.io.open_sequence(ckpt) as f:
          self.assertEqual(len(list(iter(f))), 1)

        # Simulate corrupting the first checkpoint.
        if i == 0:
          pg.io.writefile(ckpt, 'bad file')
        num_processed[example.id] = i + 1

    root_dir = os.path.join(
        tempfile.mkdtemp(),
        'per_example_checkpointer_with_corrupted_checkpoint_warm_start'
    )
    experiment = eval_test_helper.TestEvaluation()
    _ = experiment.run(
        root_dir, 'new', runner='sequential', plugins=[checkpointer],
        warm_start_from=run.output_root,
    )
    for leaf in experiment.leaf_nodes:
      self.assertEqual(leaf.progress.num_skipped, len(num_processed) - 1)
    self.assert_found_in_log(experiment, 'Failed to load checkpoint')

  def test_checkpointing_error(self):
    root_dir = os.path.join(
        tempfile.mkdtemp(),
        'per_example_checkpointer_with_checkpointing_error'
    )
    experiment = (eval_test_helper
                  .test_experiment_with_example_checkpointing_error())
    checkpointer = checkpointing.PerExampleCheckpointer('checkpoint.jsonl')
    _ = experiment.run(
        root_dir, 'new', runner='parallel', plugins=[checkpointer]
    )
    self.assert_found_in_log(experiment, 'Failed to checkpoint')


class BulkCheckpointerTest(CheckpointerTest):

  def test_checkpointing(self):
    root_dir = os.path.join(tempfile.mkdtemp(), 'test_bulk_checkpointer')
    experiment = eval_test_helper.test_experiment()
    checkpoint_filename = 'checkpoint.jsonl'
    checkpointer = checkpointing.BulkCheckpointer(checkpoint_filename)
    run = experiment.run(
        root_dir, 'new', runner='sequential', plugins=[checkpointer]
    )
    self.assertEqual(len(checkpointer._sequence_writer), 0)
    num_processed = {}
    for leaf in experiment.leaf_nodes:
      ckpt = run.output_path_for(leaf, checkpoint_filename)
      self.assertTrue(pg.io.path_exists(ckpt))
      with pg.io.open_sequence(ckpt) as f:
        self.assertEqual(
            len(list(iter(f))),
            # `eval_test_helper.test_experiment` has two TestEvaluation with
            # llm offset 0, which makes that evaluation run twice.
            leaf.progress.num_completed
            + (leaf.progress.num_failed if leaf.lm.offset == 0 else 0)
        )
      if leaf.id not in num_processed:
        self.assertEqual(leaf.progress.num_skipped, 0)
        num_processed[leaf.id] = leaf.progress.num_processed

    # Run again, should skip existing.
    _ = experiment.run(
        root_dir, 'latest', runner='sequential', plugins=[checkpointer]
    )
    self.assertEqual(len(checkpointer._sequence_writer), 0)
    for leaf in experiment.leaf_nodes:
      self.assertEqual(leaf.progress.num_skipped, num_processed[leaf.id])

  def test_checkpointing_error(self):
    root_dir = os.path.join(
        tempfile.mkdtemp(),
        'bulk_checkpointer_with_checkpointing_error'
    )
    experiment = (eval_test_helper
                  .test_experiment_with_example_checkpointing_error())
    checkpointer = checkpointing.BulkCheckpointer('checkpoint.jsonl')
    _ = experiment.run(
        root_dir, 'new', runner='parallel', plugins=[checkpointer]
    )
    self.assert_found_in_log(experiment, 'Failed to checkpoint')


if __name__ == '__main__':
  unittest.main()
